//*********************************************************
//
// CheckIntegrity:          Checks integrity between tables
//
// Author:	                Consyst SQL inc.
//
// (c) 2005 by Consyst SQL Inc.
//
//*********************************************************
using System;
using System.Collections;

namespace RepPP.Toolkit {
	/// <summary>
	/// Checks integrity between tables
	/// </summary>
	public class CheckIntegrity {
        /// <summary>Rep++ Connection</summary>
        RepPP.Connection        m_connection;

        //*********************************************************
        //
        /// <summary>
        /// ctor:                       Class constructor
        /// </summary>
        //
        //*********************************************************
		public CheckIntegrity(RepPP.Connection connection) {
            m_connection = connection;
		}

        //*********************************************************
        //
        /// <summary>
        /// FindFldUsingCol :           Find Field by one Column
        /// </summary>
        /// <param name="flds">         List of fields</param>
        /// <param name="strColName">   Column Name</param>
        /// <returns>
        /// Found field
        /// </returns>
        //
        //*********************************************************
        public RepPP.Field FindFldUsingCol(Fields flds, string strColName) {
            RepPP.Field fldRetVal = null;

            foreach (RepPP.Field fld in flds) {
                if (fld.ColName == strColName) {
                    fldRetVal = fld;
                    break;
                }
            }
            return(fldRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// BuildIsItemExistSqlCmd :        Build Sql Command to verify if item 
        ///                                 is referenced somewhere
        /// </summary>
        /// <param name="ginst">            Groupinstance to check</param>
        /// <param name="strTableName">     Table name containing the item to check</param>
        /// <param name="arrPKColList">     Reference Columns where to check</param>
        /// <param name="arrPKFldList">     Field containing the value to check. If null, the field attached to arrPKColList are used.</param>
        /// <param name="arrRetColList">    List of columns to select</param>
        /// <returns>
        /// Sql Command
        /// </returns>
        //
        //********************************************************************************
        protected virtual string BuildReferencedBySqlCmd(RepPP.GroupInstance   ginst, 
                                                         string                strTableName, 
                                                         string[]              arrPKColList, 
                                                         RepPP.Field[]         arrPKFldList, 
                                                         string[]              arrRetColList) {
            string          strRetVal;
            string          strWhereClause;
            string          strRetColList;
            int             iIndex;
            RepPP.Field     fld;

            if (arrPKFldList != null && arrPKFldList.Length != arrPKColList.Length) {
                throw new System.ApplicationException("IsReferencedBy : The primary keys list doesn't fit with the foreign primary keys list");
            }
            iIndex         = 0;
            strWhereClause = "";
            strRetColList  = String.Join(",", arrRetColList);
            foreach (string strColName in arrPKColList) {
                fld = (arrPKFldList != null) ? arrPKFldList[iIndex++] : 
                                               FindFldUsingCol(ginst.Group.Fields, strColName.ToUpper());
                if (fld == null) {
                    throw new System.ApplicationException(String.Format("IsReferencedBy : Cannot match a field from Group '{0}' with the Column : '{1}'", ginst.Group.Name, strColName));
                }
                if (strWhereClause != String.Empty) {
                    strWhereClause += " AND ";
                }
                strWhereClause += strColName + "=:" + fld.Group.Name + "." + fld.Name;
            }      
            strRetVal = String.Format("SELECT {0} FROM {1} WHERE {2} ORDER BY {0}", strRetColList, strTableName, strWhereClause);
            return(strRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// IsRefencedBy :                  Check if item in Groupinstance is referenced by one or more foreign tables
        /// </summary>
        /// <param name="ginst">            Groupinstance to check</param>
        /// <param name="strTableName">     Table name containing the item to check</param>
        /// <param name="arrPKColList">     Reference Columns where to check</param>
        /// <param name="arrPKFldList">     Field containing the value to check. If null, the field attached to arrPKColList are used.</param>
        /// <param name="arrRetColList">    List of columns to select</param>
        /// <param name="arrReferencedBy">  Array of array of referenced values</param>                            
        /// <returns>
        /// true if one or more integrities has been found
        /// </returns>
        //  SELECT arrRetColList FROM strTableName WHERE arrPKColList = :GINST.arrPKFldList
        //  SELECT QUNAME FROM SYSTQUEST WHERE APPSYSTNAME=:GINST.APPSYSTNAME AND COLNAME=:GINST.COLNAME
        //
        //********************************************************************************
        public virtual bool IsReferencedBy(RepPP.GroupInstance   ginst, 
                                           string                strTableName, 
                                           string[]              arrPKColList, 
                                           RepPP.Field[]         arrPKFldList,
                                           string[]              arrRetColList,
                                           out string[][]        arrReferencedBy) {
            RepPP.SqlCursor         sqlCursor   = null;
            ArrayList               arrColList;
            string[]                arrColValue;
            string                  strSqlCmd;
            bool                    bRetVal;
            int                     iColCount;
            int                     i;            

            arrReferencedBy = null;
            strSqlCmd       = BuildReferencedBySqlCmd(ginst, 
                                                      strTableName, 
                                                      arrPKColList, 
                                                      arrPKFldList,
                                                      arrRetColList);
            if (m_connection.Execute(strSqlCmd, RepPP.SqlOptimization.sdOptMultiRow, ref sqlCursor) != 0) {
                throw new System.ApplicationException("CheckIntegrityUI - IsReferencedBy : Unable to execute SQL Command, check your parameters");
            }
            using(sqlCursor) {                
                arrColList   = new ArrayList();
                iColCount    = arrRetColList.Length;
                while (sqlCursor.Next() == 0) {
                    arrColValue = new string[iColCount];
                    for (i = 0; i < iColCount; i++) {
                        arrColValue[i] = sqlCursor.GetColumnValue(i);
                   }
                   arrColList.Add(arrColValue);
                }
            }
            if (arrColList.Count != 0) {
                arrReferencedBy = new string[arrColList.Count][];
                arrColList.CopyTo(arrReferencedBy);
                bRetVal = true;
            } else {
                bRetVal = false;
            }
            return(bRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// IsRefencedBy :                  Check if item in Groupinstance is referenced by one or more foreign tables
        /// </summary>
        /// <param name="ginst">            Groupinstance to check</param>
        /// <param name="strTableName">     Table name containing the item to check</param>
        /// <param name="arrPKColList">     Reference Columns where to check</param>
        /// <param name="arrRetColList">    List of columns to select</param>
        /// <param name="arrReferencedBy">  Array of array of referenced values</param>                            
        /// <returns>
        /// true if one or more integrities has been found
        /// </returns>
        //  SELECT arrRetColList FROM strTableName WHERE arrPKColList = :GINST.arrPKFldList
        //  SELECT QUNAME FROM SYSTQUEST WHERE APPSYSTNAME=:GINST.APPSYSTNAME AND COLNAME=:GINST.COLNAME
        //
        //********************************************************************************
        public virtual bool IsReferencedBy(RepPP.GroupInstance   ginst, 
                                           string                strTableName, 
                                           string[]              arrPKColList, 
                                           string[]              arrRetColList,
                                           out string[][]        arrReferencedBy) {
            return(IsReferencedBy(ginst, strTableName, arrPKColList, null, out arrReferencedBy));
        }

        //********************************************************************************
        //
        /// <summary>
        /// BuildIsItemExistSqlCmd :        Build Sql Command to verify if an item is already defined
        /// </summary>
        /// <param name="ginst">            Groupinstance to check</param>
        /// <param name="strFKTableName">   Name of the table where we check for item existance</param>
        /// <param name="arrFKTablePK">     Primary key of the foreign table.</param>
        /// <param name="arrFldToCheck">    Field where to check in case</param>
        /// <returns>
        /// Sql Command
        /// </returns>
        //
        //********************************************************************************
        protected virtual string BuildIsItemExistSqlCmd(RepPP.GroupInstance   ginst, 
                                                      string                  strFKTableName, 
                                                      string[]                arrFKTablePK,
                                                      RepPP.Field[]           arrFldToCheck) {
            RepPP.Field fld;
            string      strRetVal;
            string      strWhereClause;
            int         iIndex;

            if (arrFKTablePK.Length != arrFldToCheck.Length) {
                throw new System.ApplicationException("IsItemExist : The primary keys list doesn't fit with the foreign primary keys list");
            }
            strWhereClause = "";
            iIndex         = 0;
            foreach (string strColName in arrFKTablePK) {
                fld = arrFldToCheck[iIndex++];
                if (strWhereClause != String.Empty) {
                    strWhereClause += " AND ";
                }
                strWhereClause += strColName + "=:" + fld.Group.Name + "." + fld.Name;
            }
            strRetVal = string.Format("SELECT {0} FROM {1} WHERE {2} FOR UPDATE",
                                      String.Join(",", arrFKTablePK),
                                      strFKTableName,
                                      strWhereClause);
            return(strRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// AddPKToArray:                   Build a list of PK of a group (with ancestors)
        /// </summary>
        /// <param name="arr">              Array to fill</param>
        /// <param name="cgrp">             Card group</param>
        /// <param name="grp">              Group containing the PK.</param>
        //
        //********************************************************************************
        private void AddPKToArray(ArrayList arr, RepPP.CardGroup cgrp, RepPP.Group grp) {
            if (cgrp != null) {
                cgrp = cgrp.Parent as CardGroup;
                if (cgrp != null) {
                    AddPKToArray(arr, cgrp, cgrp.Group);
                }
            }
            foreach (RepPP.Field fld in grp.Fields) {
                if (fld.IsPrimaryKey) {
                    arr.Add(fld);
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// IsItemExist :                   Check if item in Groupinstance is 
        ///                                 exist (Referencing)
        /// </summary>
        /// <param name="ginst">            Groupinstance to check</param>
        /// <param name="strFKTableName">   Foreign Table name where to check</param>
        /// <param name="arrFKTablePK">     Primary key of the foreign table.</param>
        /// <param name="arrFldToCheck">    Field where to check in case. If null, the primary key is used</param>
        /// <returns>
        /// true if one or more reference has been found
        /// </returns>
        // SELECT GINST.arrFldToCheck FROM strFKTableName WHERE arrFKTablePK=:GINST.arrFldToCheck
        //
        //********************************************************************************
        public virtual bool IsItemExist(RepPP.GroupInstance     ginst, 
                                        string                  strFKTableName, 
                                        string[]                arrFKTablePK,
                                        RepPP.Field[]           arrFldToCheck) {
            bool                bRetVal     = false;
            RepPP.SqlCursor     sqlCursor   = null;
            ArrayList           arr;
            string              strSqlCmd;

            if (arrFldToCheck == null) {
                arr  = new ArrayList();
                AddPKToArray(arr, ginst.CardGroup, ginst.Group);
                arrFldToCheck = new RepPP.Field[arr.Count];
                arr.CopyTo(arrFldToCheck);
            }
            strSqlCmd = BuildIsItemExistSqlCmd(ginst, strFKTableName, arrFKTablePK, arrFldToCheck);
            if (m_connection.Execute(strSqlCmd, RepPP.SqlOptimization.sdOptMultiRow, ref sqlCursor) != 0) {
                throw new System.ApplicationException("IsItemExist : Unable to execute SQL Command, check your parameters");
            }
            using(sqlCursor) {                
                while (sqlCursor.Next() == 0) {
                    bRetVal = true;
                }
            }
            return(bRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// IsItemExist :                   Check if item in Groupinstance is 
        ///                                 exist (Referencing)
        /// </summary>
        /// <param name="ginst">            Groupinstance to check</param>
        /// <param name="strFKTableName">   Foreign Table name where to check</param>
        /// <param name="arrFKTablePK">     Primary key of the foreign table.</param>
        /// <returns>
        /// true if one or more reference has been found
        /// </returns>
        // SELECT COUNT(*) FROM strFKTableName WHERE arrFKTablePK=:Field attached to arrFKTablePK
        //
        //********************************************************************************
        public virtual bool IsItemExist(RepPP.GroupInstance     ginst, 
                                        string                  strFKTableName, 
                                        string[]                arrFKTablePK) {
            return(IsItemExist(ginst, strFKTableName, arrFKTablePK, null));
        }

        //********************************************************************************
        //
        /// <summary>
        /// IsPKExist :                     Check if the PK exist
        /// </summary>
        /// <param name="ginst">            Groupinstance to check</param>
        /// <returns>
        /// true if one or more reference has been found
        /// </returns>
        //
        //********************************************************************************
        public virtual bool IsPKExist(RepPP.GroupInstance ginst) {
            bool            bRetVal;
            ArrayList       arr;
            string[]        arrTablePK;
            int             iIndex;

            arr = new ArrayList();
            AddPKToArray(arr, ginst.CardGroup, ginst.Group);
            arrTablePK = new string[arr.Count];
            iIndex     = 0;
            foreach (RepPP.Field fld in arr) {
                arrTablePK[iIndex++] = fld.ColName;
            }
            bRetVal = IsItemExist(ginst, ginst.Group.TableName, arrTablePK);
            return(bRetVal);
        }
	} // Class CheckIntegrity
} // Namespace
