using System;
using System.Data;
using System.Collections;
using System.ComponentModel;
using System.Runtime.Serialization;
using RepPP;

namespace RepPP.Toolkit {
    //********************************************************************************
    //
	/// <summary>
	/// DataSource interface to permit access to group instance to data binding controls
	/// </summary>
	//
    //********************************************************************************
	[Serializable]
	public class DataSourceGInst : MarshalByValueComponent, IList, IBindingList, IComparer, ITypedList, ISerializable, CrdGrpControlUtil.ICrdGrpControl {
	
	    /// <summary>Determines if description or code is displayed for choice fields</summary>
	    public enum ShowDescriptionE {
	        /// <summary>Never show the description of choice field</summary>
	        Never           = 0,
	        /// <summary>Show description but not when the user edit it</summary>
	        NotOnEdit       = 1,
	        /// <summary>Show the description all the times</summary>
	        Always          = 2
	    };
        /// <summary>The types of an item change</summary>
        public enum ItemChangedTypeE {
            /// <summary>Item changed </summary>
            ItemChanged     = 0,
            /// <summary>Item deleted </summary>
            ItemDeleted     = 1,
            /// <summary>Item inserted </summary>
            ItemInserted    = 2
        };
	    
	    /// <summary>Group control utility class</summary>
	    protected CrdGrpControlUtil             m_cgrpControlUtil;   // 
	    private ArrayList                       m_arrayList;        // Array of rows (lines)
	    private Field[]                         m_arrField;         // List of visible fields used as columns
		private bool                            m_bAllowNew;        // true to allow insertion in the source
		private bool                            m_bAllowEdit;       // true to allow modification in the source
		private bool                            m_bAllowRemove;     // true to remove a line from the source
		private bool                            m_bAllowSort;       // true to allow sort
		private int                             m_iSortedColumn;    // Column use to sort the list (-1 if none)
		private ListSortDirection               m_eSortDirection;   // Direction of the sort
		private Hashtable                       m_hashCurFldErrList;// Current list of field in errors
		private ShowDescriptionE                m_eShowDesc;        // Show description for choice
		private string                          m_strWhereClause;   // WHERE clause for autofill
		private bool                            m_bAutoFill;        // Autofill the datasource if true
        private bool                            m_bSortGroupInst;   // Sort the actutual GroupInstance if true
        private GroupInstance                   m_ginstAutoFill;    // Autofill group instance if any
		internal DataRowGInst                   m_ginstRowNewLine;  // New line if in process.
		/// <summary>Event trigger when the data source changed</summary>
        public event ListChangedEventHandler    ListChanged;        // Event

        //********************************************************************************
        //
        /// <summary>
        /// ctor:                           Private constructor
        /// </summary>
        /// <param name="parent">           Control embedding this class. Null except if the DataSourceGInst is also embedded.</param>
		/// <param name="strCardName">      Rep++ Card name</param>
		/// <param name="strCardGroupName"> Rep++ CardGroup name</param>
		/// <param name="iLineCount">       Number of lines in the source</param>
        /// <param name="bNoRefCount">      true to cancel reference counting (component dispose is not always called)</param>
        //
        //********************************************************************************
		private DataSourceGInst(object parent, string strCardName, string strCardGroupName, int iLineCount, bool bNoRefCount) {
		    m_cgrpControlUtil           = new CrdGrpControlUtil((parent == null) ? this : parent,
		                                                        false,          // Not always used with resource dispenser
                                                                bNoRefCount);   // The Dispose is not called by the Page default implementation. Must not use ref counting
		    m_arrayList                 = new ArrayList(iLineCount);
		    Fields                      = null;
		    m_bAllowNew                 = true;
		    m_bAllowEdit                = true;
		    m_bAllowRemove              = true;
		    m_bAllowSort                = true;
            m_bSortGroupInst            = true;
		    m_eShowDesc                 = ShowDescriptionE.Always;
		    m_iSortedColumn             = -1;
		    m_hashCurFldErrList         = null;
		    m_eSortDirection            = ListSortDirection.Ascending;
            m_cgrpControlUtil.CardGroupFullName = new CrdGrpFullName(strCardName, strCardGroupName);
		    m_strWhereClause            = "";
		    m_bAutoFill                 = false;
		    m_ginstAutoFill             = null;
		}

        //********************************************************************************
        //
        /// <summary>
        /// ctor:                       Build the class from a group instance. The source will contains
        ///                             all the row of the instance
        /// </summary>
		/// <param name="info">         Serialization object</param>
		/// <param name="context">      Serialization context</param>
        //
        //********************************************************************************
		protected DataSourceGInst(SerializationInfo info, StreamingContext context) : this(null, "", "", 0, true) {
		    string              strCardName;
		    string              strCardGroupName;
		    
		    strCardName                         = info.GetString("CardName");
		    strCardGroupName                    = info.GetString("CardGroupName");
		    m_eShowDesc                         = (ShowDescriptionE)info.GetInt32("ShowDesc");
    		m_bAllowNew                         = info.GetBoolean("AllowNew");
	    	m_bAllowEdit                        = info.GetBoolean("AllowEdit");
		    m_bAllowRemove                      = info.GetBoolean("AllowRemove");
		    m_bAllowSort                        = info.GetBoolean("AllowSort");
            m_bSortGroupInst                    = info.GetBoolean("SortGroupInst");
            m_cgrpControlUtil.CardGroupFullName = new CrdGrpFullName(strCardName, strCardGroupName);
		}

        //********************************************************************************
        //
        /// <summary>
        /// ctor:                       Build the class from a group. The source will be empty
        /// </summary>
        //
        //********************************************************************************
		public DataSourceGInst() : this(null, "", "", 0, true) {
		}

        //********************************************************************************
        //
        /// <summary>
        /// ctor:                       Build the class from a group. The source will be empty
        /// </summary>
        /// <param name="parent">       Control embedding this class. Null except if the DataSourceGInst is also embedded.</param>
        /// <param name="bNoRefCount">  true to cancel reference counting (component dispose is not always called)</param>
        /// 
        //
        //********************************************************************************
		public DataSourceGInst(object parent, bool bNoRefCount) : this(parent, "", "", 0, bNoRefCount) {
		}

        //********************************************************************************
        //
        /// <summary>
        /// ctor:                       Build the class from a group. The source will be empty
        /// </summary>
        //
        //********************************************************************************
		public DataSourceGInst(System.ComponentModel.IContainer container) : this(null, "", "", 0, true) {
			container.Add(this);
		}

        //********************************************************************************
        //
        /// <summary>
        /// ctor:                       Build the class from a group. The source will be empty
        /// </summary>
		/// <param name="cgrp">          Rep++ CardGroup</param>
        //
        //********************************************************************************
		public DataSourceGInst(CardGroup cgrp) : this(null, cgrp.Card.Name, cgrp.Group.Name, 0, true) {
		}

        //********************************************************************************
        //
        /// <summary>
        /// ctor:                       Build the class from a group instance. The source will contains
        ///                             all the row of the instance
        /// </summary>
		/// <param name="ginst">        Rep++ Group instance</param>
        //
        //********************************************************************************
		public DataSourceGInst(GroupInstance ginst) : this(null, ginst.CardInstance.Card.Name, ginst.Group.Name, ginst.UndelLineCount, true) {
		    GroupInstance = ginst;
		}

        //********************************************************************************
        //
        /// <summary>
        /// Dispose:                    Dispose the object
        /// </summary>
		/// <param name="bDisposing">   true if dispose, false if GC</param>
        //
        //********************************************************************************
        protected override void Dispose(bool bDisposing) {            
            if (m_ginstAutoFill != null) {
                m_ginstAutoFill.Delete();
                m_ginstAutoFill = null;
            }
            if (m_cgrpControlUtil != null) {
                m_cgrpControlUtil.Dispose();
                m_cgrpControlUtil = null;
            }
            base.Dispose(bDisposing);
        }

        //********************************************************************************
        //
		/// <summary>
		/// RepPPReInit:        Called when Rep++ is reinitialized in design mode
		/// </summary>
		//
        //********************************************************************************
        public void RepPPReInit() {
        }

        //********************************************************************************
        //
		/// <summary>
		/// RepPPReset:         Called when Rep++ is reset in design mode
		/// </summary>
		//
        //********************************************************************************
        public void RepPPReset() {
        }

        //********************************************************************************
        //
		/// <summary>
		/// GetObjectData:              Serialized the object
		/// </summary>
		/// <param name="info">         Information to serialize</param>
		/// <param name="context">      Context</param>
        //
        //********************************************************************************
		public virtual void GetObjectData(SerializationInfo info, StreamingContext context) { // ISerializable
		    info.AddValue("CardName",       m_cgrpControlUtil.CardGroupFullName.CardName);
		    info.AddValue("CardGroupName",  m_cgrpControlUtil.CardGroupFullName.CardGroupName);
		    info.AddValue("ShowDesc",       (int)m_eShowDesc);
		    info.AddValue("AllowNew",       m_bAllowNew);
		    info.AddValue("AllowEdit",      m_bAllowEdit);
		    info.AddValue("AllowRemove",    m_bAllowRemove);
		    info.AddValue("AllowSort",      m_bAllowSort);
		    info.AddValue("SortGroupInst",  m_bSortGroupInst);
		}

        //********************************************************************************
        //
		/// <summary>
		/// Fields:                     Array of fields
		/// </summary>
        //
        //********************************************************************************
		internal Field[] Fields {
		    get {
                Field[] arrRetFlds = null;

		        if (m_arrField == null) {
		            arrRetFlds = FillStruct();
                    if (m_cgrpControlUtil != null && !m_cgrpControlUtil.IsDesignMode) {
                       m_arrField = arrRetFlds;
                    }
		        } else {
                    arrRetFlds = m_arrField;              
                }
		        return(arrRetFlds);
		    }
		    set {
               if (m_cgrpControlUtil != null && !m_cgrpControlUtil.IsDesignMode) {
		            m_arrField = value;
               }
		    }
		}

        //********************************************************************************
        //
        /// <summary>
        /// TriggerComponentChange:     Trigger the event indicating a component property has changed
        /// </summary>
        //
        //********************************************************************************
        public virtual void TriggerComponentChange() {
            if (m_cgrpControlUtil != null) {
                m_cgrpControlUtil.TriggerComponentChange();
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// Group:              Attached group
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
		[Description("Group which will be represented by this Data Source")]
		[DesignerSerializationVisibilityAttribute(DesignerSerializationVisibility.Hidden)]
        public virtual Group Group {
            get {
                return(m_cgrpControlUtil.Group);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// CardGroupFullName:                  Name of the CardGroup attached to the source
        /// </summary>
        //
        //********************************************************************************
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
        [Editor(typeof(CrdGrpControlUtil.CrdGrpControlEditor), typeof(System.Drawing.Design.UITypeEditor))]
        [Browsable(true)]
        [Bindable(true)]
        [Category("Data")]
        [Description("Name of the CardGroup attached to the source")]
		public virtual CrdGrpFullName CardGroupFullName {
		    get {
		        return(m_cgrpControlUtil.CardGroupFullName);
		    }
		    set {
		        m_cgrpControlUtil.CardGroupFullName = value;
		        Fields = null;
		    }
		}

        //********************************************************************************
        //
        /// <summary>
        /// WhereClause:            WHERE clause to use to autofill the datasource
        /// </summary>
        //
        //********************************************************************************
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
        [Browsable(true)]
        [Bindable(true)]
        [Category("Data")]
        [Description("WHERE clause use to auto-fill the data source")]
		public virtual string WhereClause {
		    get {
		        return(m_strWhereClause);
		    }
		    set {
		        m_strWhereClause = (value == null) ? "" : value;
		    }
		}

        //********************************************************************************
        //
        /// <summary>
        /// AutoFill:               Use to auto-fill the data source
        /// </summary>
        //
        //********************************************************************************
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
        [Browsable(true)]
        [Bindable(true)]
        [Category("Data")]
        [Description("Auto fill the data source")]
		public virtual bool AutoFill {
		    get {
		        return(m_bAutoFill);
		    }
		    set {
		        m_bAutoFill = value;
		    }
		}

        //********************************************************************************
        //
        /// <summary>
        /// AutoAssociateGroupInstance: Automaticaly associate a group instance to the data source
        /// </summary>
        //
        //********************************************************************************
        public virtual void AutoAssociateGroupInstance(CardInstance cinst) {
            Card            crd;
            CardGroup       cgrp;
            GroupInstance   ginstRoot;
            
            cgrp = m_cgrpControlUtil.CardGroup;
            if (cinst == null) {
                crd         = cgrp.Card;
                ginstRoot   = crd.ChildCardGroups[0].Group.ActiveGroupInstance;
                if (ginstRoot != null && ginstRoot.CardInstance != null && ginstRoot.CardInstance.Card == crd) {
                    cinst = ginstRoot.CardInstance;
                }
            }
            if (cinst != null) {
                GroupInstance = cinst.GetCurrentGroupInstance(cgrp);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// GroupInstance:      Attached group instance
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
		[Description("GroupInstance represented by this Data Source")]
        [DesignerSerializationVisibilityAttribute(DesignerSerializationVisibility.Hidden)]
        public virtual GroupInstance GroupInstance {
            get {
                CheckForAutoFill();
                return((m_ginstAutoFill != null) ? m_ginstAutoFill : m_cgrpControlUtil.GroupInstance);
            }
            set {
                if (value == null || m_cgrpControlUtil.GroupInstance != value) {
                    if (value != null) {
                        if (value.Group != m_cgrpControlUtil.Group) {
                            throw new ArgumentException("The group instance doesn't point to the same group");
                        }
                    }
                    if (m_ginstAutoFill != null) {
                        m_ginstAutoFill.Delete();
                        m_ginstAutoFill = null;
                        AutoFill        = false;
                    }
                    m_cgrpControlUtil.GroupInstance = value;
                    FillRow();
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// Application:            Gets the Rep++ application
        /// </summary>
        //
        //********************************************************************************
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Browsable(false)]
		public Application Application {
		    get {
		        return(m_cgrpControlUtil.Application);
		    }
		}

        //********************************************************************************
        //
		/// <summary>
		/// FillStruct:                 Fill structure
		/// </summary>
        //
        //********************************************************************************
        private Field[] FillStruct() {

		    int     iVisibleField;
		    Field[] arrRetFlds      = null;
		    Group   grp;
		    
		    m_arrayList.Clear();
		    grp = Group;
		    if (grp != null) {
		        iVisibleField   = 0;
		        foreach (Field fld in grp.Fields) {
		            if (fld.IsVisible) {
		                iVisibleField++;
		            }
		        }
		        arrRetFlds    = new Field[iVisibleField];
		        iVisibleField = 0;
		        foreach (Field fld in grp.Fields) {
		            if (fld.IsVisible) {
		                arrRetFlds[iVisibleField++] = fld;
		            }
		        }
		    }
            return (arrRetFlds);
        }

        //********************************************************************************
        //
		/// <summary>
		/// FillRow:            Fill the row with the content of the group instance
		/// </summary>
        //
        //********************************************************************************
        private void FillRow() {
            int                 iLineCount;
            int                 iUndelLineCount;
            DataRowGInst        ginstRow;
            GroupInstance       ginst;
            Field[]             arrFld;
            
            arrFld = Fields;    // Ensure the field array is filled.
            ginst = GroupInstance;
            m_arrayList.Clear();
            if (ginst != null) {
                iUndelLineCount         = ginst.UndelLineCount;
                iLineCount              = ginst.LineCount;
                m_arrayList.Capacity = iUndelLineCount;
                for (int iIndex = 0; iIndex < iLineCount; iIndex++) {
                    if (ginst.ThisLineState(iIndex) != LineState.sdLineDeleted) {
                        ginstRow = new DataRowGInst(this, ginst.LineBookmark(iIndex));
                        m_arrayList.Add(ginstRow);
                    }
                }
            }
            OnListChanged(new ListChangedEventArgs(ListChangedType.Reset, 0));
        }

        //********************************************************************************
        //
		/// <summary>
		/// CreateGInst:          Create an autofill group instance
		/// </summary>
        //
        //********************************************************************************
        private void CreateGInst() {
            string  strWhereClause;
            string  strSelect = null;
            string  strUpdate = null;
            string  strInsert = null;
            string  strDelete = null;
            Group   grp;
            
            if (m_ginstAutoFill != null) {
                m_ginstAutoFill.Delete();
                m_ginstAutoFill = null;
            }
            grp             = Group;
            strWhereClause  = WhereClause;
            grp.BuildSqlCommand("", ref strSelect, ref strUpdate, ref strInsert, ref strDelete);
            strSelect       = grp.Application.Tool.SetSqlWhere(strSelect, strWhereClause);
            m_ginstAutoFill = grp.GroupInstances.Create(GroupType.sdHorizontal);
            m_ginstAutoFill.ReadFromDb(strSelect);
            FillRow();
        }

        //********************************************************************************
        //
		/// <summary>
		/// CheckForAutoFill:       Check if we must auto-fill
		/// </summary>
        //
        //********************************************************************************
        private void CheckForAutoFill() {
            if (AutoFill && m_ginstAutoFill == null) {
                CreateGInst();
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// GetFieldNameIndex:          Gets the index of a field name in the list of columns
        /// </summary>
		/// <param name="strName">      Field name</param>
		/// <returns>
		/// Index or -1 if not found
		/// </returns>
        //
        //********************************************************************************
		private int GetFieldNameIndex(string strName) {
		    int     iRetVal = -1;
		    int     iIndex;
		    Field[] arrFld;
		    
		    iIndex = 0;
		    arrFld = Fields;
		    if (arrFld != null) {
		        foreach (Field fld in arrFld) {
		            if (String.Compare(fld.Name, strName, true) == 0) {
		                iRetVal = iIndex;
		                break;
		            } else {
		                iIndex++;
		            }
		        }
		    }
		    return(iRetVal);
		}

        //********************************************************************************
        //
        /// <summary>
        /// Application_FldOnError:     Intercept field error
        /// </summary>
        /// <param name="sender">       Sender object</param>
        /// <param name="e">            Event parameter</param>
        //
        //********************************************************************************
        private void Application_FldOnError(object sender, FieldGenEventArgs e) {
            if (m_hashCurFldErrList != null) {
                if (e.Field.Group == Group) {
                    m_hashCurFldErrList.Add(e.Field.Name, e.ErrorText);
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// IsLineValid:                    Check if line is valid. Returns the list of errors
        ///                                 in hashFldInError
        /// </summary>
        /// <param name="hashFldInError">   Hashtable to fill with the error</param>
        /// <returns>
        /// true:                           Line is valid
        /// false:                          Line is not
        /// </returns>
        //
        //********************************************************************************
        internal bool IsLineValid(Hashtable hashFldInError) {
            bool    bRetVal = true;
            Group   grp;
            Field[] arrFld;
            
		    grp = Group;
		    grp.Application.FldOnError += new FldOnErrorEventHandler(Application_FldOnError);
            m_hashCurFldErrList = hashFldInError;
            m_hashCurFldErrList.Clear();
            arrFld = Fields;
            foreach (Field fld in arrFld) {
                if (fld.Validate() != 0) {
                    bRetVal = false;
                }
            }
            m_hashCurFldErrList = null;
		    grp.Application.FldOnError -= new FldOnErrorEventHandler(Application_FldOnError);
            return(bRetVal);
        }		

        //********************************************************************************
        //
        /// <summary>
        /// OnListChanged:      Trigger the ListChanged event
        /// </summary>
        /// <param name="e">    Event parameter</param>
        //
        //********************************************************************************
        public void OnListChanged(ListChangedEventArgs e) {
            if (ListChanged != null) {
                ListChanged(this, e);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// ShowDescription:        Show description of choice if true, code if false
        /// </summary>
        //
        //********************************************************************************
		[Browsable(true)]
		[Category("Behavior")]
		[Description("How to display a choice value")]
        public ShowDescriptionE ShowDescription {
            get {
                return(m_eShowDesc);
            }
            set {
                m_eShowDesc = value;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// Sort:               Returns the colums use for sort (or "" if none)
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
        [DesignerSerializationVisibilityAttribute(DesignerSerializationVisibility.Hidden)]
        public string Sort {
            get {
                Field[] arrFld;
                string  strRetVal;
                
                arrFld    = Fields;
                strRetVal = (m_iSortedColumn == -1) ? 
                            "" : 
                            arrFld[m_iSortedColumn].Name;
                return(strRetVal);
            }
            set {
                int                 iIndex;
                ListSortDirection   eDir;
                object[]            arrSortParam;
                
                if (!m_bAllowSort) {
                    throw new InvalidOperationException("Sort not allowed");
                }
                if (value != "") {
                    if (value.IndexOf(',') != -1) {
                        throw new InvalidOperationException("Sort accept only simple field name");
                    }
                    value = value.ToUpper();
                    if (value.EndsWith(" ASC")) {
                        eDir  = ListSortDirection.Ascending;
                        value = value.Substring(0, value.Length - 4).Trim();
                    } else if (value.EndsWith(" DESC")) {
                        eDir  = ListSortDirection.Descending;
                        value = value.Substring(0, value.Length - 5).Trim();
                    } else {
                        eDir  = ListSortDirection.Ascending;
                    }
                    iIndex = GetFieldNameIndex(value);
                    if (iIndex == -1) {
                        throw new ArgumentException("Column '" + value + "' is not a valid column name");
                    }
                    m_iSortedColumn  = iIndex;
                    m_eSortDirection = eDir;
                }
                if (SortGroupInst) {
                    arrSortParam    = new object[2];
                    arrSortParam[0] = Fields[m_iSortedColumn].Name;
                    arrSortParam[1] = m_eSortDirection;
                    GroupInstance.Sort(arrSortParam);
                    FillRow();
                } else {
                    if (Count == 0) {
                       FillRow();
                    }
                    m_arrayList.Sort(this);
                }
                OnListChanged(new ListChangedEventArgs(ListChangedType.Reset, 0));
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// IsReadOnly:         Returns if the list is read-only
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
        public bool IsReadOnly {            // IList
            get {
                return(!m_bAllowNew && !m_bAllowEdit && !m_bAllowRemove);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        ///  Default indexer
        /// </summary>
        //
        //********************************************************************************
        public object this[int iIndex] {     // IList
            get {
                CheckForAutoFill();
                return(m_arrayList[iIndex]);
            }
            set {
                throw new InvalidOperationException("The Item property is read-only");
            }
        }

        //********************************************************************************
        //
        /// <summary>
        ///  RemoveAt:              Remove the item from the list
        /// </summary>
        /// <param name="iIndex">    Index of the item to remove</param>
        //
        //********************************************************************************
        public void RemoveAt(int iIndex) {   // IList
            GroupInstance       ginst;
            DataRowGInst        ginstRow;
            int                 iLineToDel;
            
            if (!m_bAllowRemove) {
                throw new InvalidOperationException("Rows cannot be deleted");
            }
            ginstRow = m_arrayList[iIndex] as DataRowGInst;
            if (ginstRow == null) {
                throw new ArgumentException("Bad argument type");
            }
            ginst      = GroupInstance;
            iLineToDel = ginst.LineFromBookmark(ginstRow.Bookmark);
            if (iLineToDel == -1) {
                throw new ArgumentException("Bad index");
            }
            if (iLineToDel == ginst.CurrentLine) {
                ginst.DeleteLine();
            } else {
                ginst.SaveLinePos();
                try {
                    ginst.SelectLine(iLineToDel, false, false);
                    ginst.DeleteLine();
                } finally {
                    ginst.RestoreLinePos(true);
                }
            }
            m_arrayList.RemoveAt(iIndex);
            OnListChanged(new ListChangedEventArgs(ListChangedType.ItemDeleted, iIndex));
        }

        //********************************************************************************
        //
        /// <summary>
        ///  Insert:                Insert an item in the list
        /// </summary>
        /// <param name="index">    Position at which to insert the item</param>
        /// <param name="value">    Item to insert</param>
        //
        //********************************************************************************
        public void Insert(int index, object value) {   // IList
            throw new InvalidOperationException("The list is read-only. Use to AddNew method.");
        }

        //********************************************************************************
        //
        /// <summary>
        /// Remove:                 Remove an item from the list
        /// </summary>
        /// <param name="value">    Item to remove</param>
        //
        //********************************************************************************
        public void Remove(object value) {  // IList
            RemoveAt(m_arrayList.IndexOf(value));
        }

        //********************************************************************************
        //
        /// <summary>
        ///  Contains:              Check if the list contains the specified object.
        /// </summary>
        /// <param name="value">    Object to check</param>
        /// <returns>
        /// true if yes, false if not.
        /// </returns>
        //
        //********************************************************************************
        public bool Contains(object value) {    // IList
            CheckForAutoFill();
            return(m_arrayList.Contains(value));
        }

        //********************************************************************************
        //
        /// <summary>
        ///  Clear:                 Removes all items from the list
        /// </summary>
        //
        //********************************************************************************
        public void Clear() {   // IList
            GroupInstance   ginst;
            
            ginst = GroupInstance;
            if (ginst != null) {
                ginst.Empty(GroupInstDelFlag.sdDelKey);
            }
            m_arrayList.Clear();
            OnListChanged(new ListChangedEventArgs(ListChangedType.Reset, 0));
        }

        //********************************************************************************
        //
        /// <summary>
        ///  IndexOf:               Returns the index of the object in the list
        /// </summary>
        /// <param name="value">    Object</param>
        /// <returns>
        /// Index or -1 if not found
        /// </returns>
        //
        //********************************************************************************
        public int IndexOf(object value) {      // IList
            CheckForAutoFill();
            return(m_arrayList.IndexOf(value));
        }

        //********************************************************************************
        //
        /// <summary>
        ///  Add:                   Adds a new item to the list. Use to AddNew method instead.
        /// </summary>
        /// <param name="value">    Object to add.</param>
        //
        //********************************************************************************
        public int Add(object value) {      // IList
            throw new InvalidOperationException("The list is read-only. Use to AddNew method.");
        }

        //********************************************************************************
        //
        /// <summary>
        ///  IsFixedSize:           Returns if the list has a fixed size.
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
        public bool IsFixedSize {   // IList
            get {
                return(!(m_bAllowNew || m_bAllowRemove));
            }
        }

        //********************************************************************************
        //
        /// <summary>
        ///  IsSynchronized:        Returns false. This list is not thread safe
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
        public bool IsSynchronized {    // ICollection
            get {
                return(false);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        ///  Count:                 Returns the number of lines (rows) in the source
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
        public int Count {              // ICollection
            get {
                CheckForAutoFill();
                return(m_arrayList.Count);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        ///  CopyTo:                Copy the content of the list to the specified array
        /// </summary>
        /// <param name="array">    Array to copy to</param>
        /// <param name="iIndex">   Index to start with</param>
        //
        //********************************************************************************
        public void CopyTo(Array array, int iIndex) {    // ICollection
            CheckForAutoFill();
            m_arrayList.CopyTo(array, iIndex);
        }

        //********************************************************************************
        //
        /// <summary>
        ///  SyncRoot:              Returns the synchronize root. This list is not thread safe.
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
        public object SyncRoot {        // ICollection
            get {
                return(this);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        ///  GetEnumerator:         Returns an enumerator for the list
        /// </summary>
        /// <returns>
        /// Enumerator for the list
        /// </returns>
        //
        //********************************************************************************
        public IEnumerator GetEnumerator() {    // IEnumerator
            CheckForAutoFill();
            return(m_arrayList.GetEnumerator());
        }

        //********************************************************************************
        //
        /// <summary>
        /// CompareValues:          Compare two objects value
        /// </summary>
        /// <param name="x">        First object</param>
        /// <param name="y">        Second object</param>
        /// <returns>
        /// -1 (x .lt. y), 0 (x == y), 1 (x .gt. y)
        /// </returns>
        //
        //********************************************************************************
        private int CompareValues(object x, object y) {
            int     iRetVal;
            string  strVal1;
            string  strVal2;
            
            if (x == DBNull.Value) {
                if (y == DBNull.Value) {
                    iRetVal = 0;
                } else {
                    iRetVal = -1;
                }
            } else {
                if (y == DBNull.Value) {
                    iRetVal = 1;
                } else {
                    strVal1 = x.ToString();
                    strVal2 = y.ToString();
                    iRetVal = String.Compare(strVal1, strVal2);
                }
            }
            return(iRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        ///  Compare:               Compare two objects
        /// </summary>
        /// <param name="x">        First object</param>
        /// <param name="y">        Second object</param>
        /// <returns>
        /// -1 (x .lt. y), 0 (x == y), 1 (x .gt. y)
        /// </returns>
        //
        //********************************************************************************
        public int Compare(object x, object y) {
            int                 iRetVal;
            object              objVal1;
            object              objVal2;
            
            objVal1 = (x is DataRowGInst) ? (x as DataRowGInst)[m_iSortedColumn] : x;
            objVal2 = (y is DataRowGInst) ? (y as DataRowGInst)[m_iSortedColumn] : y;
            iRetVal = CompareValues(objVal1, objVal2);
            if (m_eSortDirection == ListSortDirection.Descending) {
                iRetVal = -iRetVal;
            }
            return(iRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// AddIndex:               Adds an index for the sort.
        /// </summary>
        /// <param name="property"> Property of the index</param>
        //
        //********************************************************************************
        public void AddIndex(PropertyDescriptor property) {     // IBindingList
            // Do nothing. Not supported.
        }

        //********************************************************************************
        //
        /// <summary>
        /// AllowNew:           Determines if the source allow insertion of rows
        /// </summary>
        //
        //********************************************************************************
		[Browsable(true)]
		[Category("Behavior")]
		[Description("Allow the insertion of new rows")]
        public bool AllowNew {  // IBindingList
            get {
                bool    bRetVal;
                
                if (DesignMode) {
                    bRetVal = m_bAllowNew;
                } else {
                    bRetVal = m_bAllowNew && (GroupInstance != null);
                }
                return(bRetVal);
            }
            set {
                m_bAllowNew = value;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// SortGroupInst:           Determines if the sort will be performed locally 
        ///                          or using the internal list
        /// </summary>
        //
        //********************************************************************************
		[Browsable(true)]
		[Category("Behavior")]
		[Description("Sort the GroupInstance instead of the internal list")]
        public bool SortGroupInst {
            get {
                return(m_bSortGroupInst);
            }
            set {
                m_bSortGroupInst = value;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// ApplySort:                  Apply the sort
        /// </summary>
        /// <param name="property">     Property use to sort</param>
        /// <param name="direction">    Direction of the sort</param>
        //
        //********************************************************************************
        public void ApplySort(PropertyDescriptor property, ListSortDirection direction) {   // IBindingList
            string  strName;
            
            strName = (direction == ListSortDirection.Ascending) ? property.Name : property.Name + " DESC";
            Sort    = strName;
        }

        //********************************************************************************
        //
        /// <summary>
        ///  SortProperty:          Returns the property which is actually sorted
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
        public PropertyDescriptor SortProperty {    // IBindingList
            get {
                PropertyDescriptor  propRetVal;
                Field[]             arrFld;
                Field               fld;
                
                if (m_iSortedColumn == -1) {
                    propRetVal = null;
                } else {
                    arrFld     = Fields;
                    fld        = arrFld[m_iSortedColumn];
                    propRetVal = new DataRowGInstPropertyDescriptor(fld.Name,
                                                                    fld.Prompt,
                                                                    m_iSortedColumn,
                                                                    typeof(string),
                                                                    !fld.IsModifiable);
                }
                return(propRetVal);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        ///  Find:                      Find a raw using the key of a property
        /// </summary>
        /// <param name="strColName">   Column to use for the search</param>
        /// <param name="key">          Key value</param>
        /// <returns>
        /// Index or -1 if not found
        /// </returns>
        //
        //********************************************************************************
        public int Find(string strColName, object key) {
            int     iRetVal = -1;
            int     iColIndex;
            int     iCount;
            object  objVal;

            iColIndex = GetFieldNameIndex(strColName);
            if (iColIndex == -1) {
                throw new ArgumentException("Column '" + strColName + "' does not exist in the source.");
            }
            if (iColIndex == m_iSortedColumn) {
                iRetVal = m_arrayList.BinarySearch(key, this);
            } else {
                iCount  = m_arrayList.Count;
                for (int iIndex = 0; iIndex < iCount; iIndex++) {
                    objVal = (m_arrayList[iIndex] as DataRowGInst)[iColIndex];
                    if (CompareValues(key, objVal) == 0) {
                        iRetVal = iIndex;
                        break;
                    }
                }
            }
            return(iRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        ///  Find:                  Find a raw using the key of a property
        /// </summary>
        /// <param name="property"> Property use to find the item</param>
        /// <param name="key">      Key value</param>
        /// <returns>
        /// Index
        /// </returns>
        //
        //********************************************************************************
        public int Find(PropertyDescriptor property, object key) {  // IBindingList
            return(Find(property.Name, key));
        }

        //********************************************************************************
        //
        /// <summary>
        ///  SupportsSorting:       Returns true to indicate the source support sorting
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
        public bool SupportsSorting {   // IBindingList
            get {
                return(true);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// IsSorted:           Returns true if source is sorted
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
        public bool IsSorted {  // IBindingList
            get {
                return(m_iSortedColumn != -1);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// AllowRemove:        Determines if the source allow deletion of rows
        /// </summary>
        //
        //********************************************************************************
		[Browsable(true)]
		[Category("Behavior")]
		[Description("Allow removal of rows")]
        public bool AllowRemove {   // IBindingList
            get {
                return(m_bAllowRemove);
            }
            set {
                m_bAllowRemove = value;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// SupportsSearching:      Returns true if the source support searching
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
        public bool SupportsSearching { // IBindingList
            get {
                return(true);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// SortDirection:      Returns the sort direction
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
        public ListSortDirection SortDirection {    // IBindingList
            get {
                return(m_eSortDirection);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// SupportsChangeNotification: Returns true if the source support change notification
        /// </summary>
        //
        //********************************************************************************
		[Browsable(false)]
        public bool SupportsChangeNotification {    // IBindingList
            get {
                return(true);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// RemoveSort:                 Remove a sort column
        /// </summary>
        //
        //********************************************************************************
        public void RemoveSort() {  // IBindingList
            Sort = "";
        }

        //********************************************************************************
        //
        /// <summary>
        ///  AddNew:                Adds a new row (line) to the source
        /// </summary>
        /// <returns>
        /// New line (row)
        /// </returns>
        //
        //********************************************************************************
        public object AddNew() {    // IBindingList
            int             iBmk;
            GroupInstance   ginst;
            
            ginst = GroupInstance;
            if (!m_bAllowNew || ginst == null) {
                throw new InvalidOperationException("AddNew is not allowed");
            }
            ginst.Select(false, true);
            ginst.AddBlankLine(true);
            iBmk = ginst.LineBookmark(ginst.CurrentLine);
            m_ginstRowNewLine = new DataRowGInst(this, iBmk);
            m_arrayList.Add(m_ginstRowNewLine);
            OnListChanged(new ListChangedEventArgs(ListChangedType.ItemAdded, m_arrayList.Count - 1));
            return(m_ginstRowNewLine);
        }

        //********************************************************************************
        //
        /// <summary>
        /// AllowEdit:          Determines if the source allow modification of items
        /// </summary>
        //
        //********************************************************************************
		[Browsable(true)]
		[Category("Behavior")]
		[Description("Allow edition of a row")]
        public bool AllowEdit { // IBindingList
            get {
                return(m_bAllowEdit);
            }
            set {
                m_bAllowEdit = value;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        ///  AddDuplicate:                Adds a duplicate row (line) to the source
        /// </summary>
        /// <returns>
        /// New line (row)
        /// </returns>
        //
        //********************************************************************************
        public object AddDuplicate() {
            int             iBmk;
            GroupInstance   ginst;
            
            ginst = GroupInstance;
            if (!m_bAllowNew || ginst == null) {
                throw new InvalidOperationException("AddDuplicate is not allowed");
            }
            ginst.Select(false, true);
            ginst.DuplicateLine();
            iBmk = ginst.LineBookmark(ginst.CurrentLine);
            m_ginstRowNewLine = new DataRowGInst(this, iBmk);
            m_arrayList.Add(m_ginstRowNewLine);
            OnListChanged(new ListChangedEventArgs(ListChangedType.ItemAdded, m_arrayList.Count - 1));
            return(m_ginstRowNewLine);
        }

        //********************************************************************************
        //
        /// <summary>
        /// RemoveIndex:            Remove a sort index
        /// </summary>
        /// <param name="property"> Property to remove</param>
        //
        //********************************************************************************
        public void RemoveIndex(PropertyDescriptor property) {  // IBindingList
            // Do nothing.
        }

        //********************************************************************************
        //
        /// <summary>
        /// GetItemProperties:              Returns a descriptor for each property (field)
        /// </summary>
        /// <param name="listAccessors">    List accessor (not used)</param>
        /// <returns>
        /// List of properties
        /// </returns>
        //
        //********************************************************************************
        public PropertyDescriptorCollection GetItemProperties(PropertyDescriptor[] listAccessors) { // ITypedList
            PropertyDescriptorCollection        colRetVal = null;
            DataRowGInstPropertyDescriptor[]    arrPropDesc;
            int                                 iIndex;
            Field[]                             arrFld;
            
            arrFld = Fields;
            if (arrFld == null) {
                arrPropDesc = new DataRowGInstPropertyDescriptor[0];
            } else {
                arrPropDesc = new DataRowGInstPropertyDescriptor[arrFld.Length];
                iIndex      = 0;
                foreach (Field fld in arrFld) {
                    arrPropDesc[iIndex] = new DataRowGInstPropertyDescriptor(fld.Name,
                                                                             fld.Prompt,
                                                                             iIndex,
                                                                             typeof(string),
                                                                             !fld.IsModifiable);
                    iIndex++;
                }
            }
            colRetVal = new PropertyDescriptorCollection(arrPropDesc);
            return(colRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// GetListName:                    Names use to identify the source (TableStyle)
        /// </summary>
        /// <param name="listAccessors">    List of accessor (not used)</param>
        /// <returns>
        /// Name of the datasource
        /// </returns>
        //
        //********************************************************************************
        public string GetListName(PropertyDescriptor[] listAccessors) {
            Group   grp;
            
            grp = Group;
            return((grp == null) ? "Source1" : grp.Name + "_Source");
        }

        //********************************************************************************
        //
        /// <summary>
        ///  Refresh:                 Refreshes the list of items in the list from the 
        ///                           attached GroupInstance
        /// </summary>
        //
        //********************************************************************************
        public void Refresh() {
            FillRow();
        }

        //********************************************************************************
        //
        /// <summary>
        ///  RefreshItem:                   Refreshes an item in the list from the attached group instance
        /// </summary>
        /// <param name="iItemIndex">       Item index</param>
        /// <param name="eItemChangeType">  The item change type</param>
        //
        //********************************************************************************
        public void RefreshItem(int iItemIndex, ItemChangedTypeE eItemChangeType) {
            int             iBmk;
            GroupInstance   ginst;

            switch (eItemChangeType) {
            case ItemChangedTypeE.ItemChanged:
                OnListChanged(new ListChangedEventArgs(ListChangedType.ItemChanged, iItemIndex));
                break;
            case ItemChangedTypeE.ItemDeleted:
                if (!m_bAllowRemove) {
                    throw new InvalidOperationException("Rows cannot be deleted");
                }
                m_arrayList.RemoveAt(iItemIndex);
                OnListChanged(new ListChangedEventArgs(ListChangedType.ItemDeleted, iItemIndex));
                
                break;
            case ItemChangedTypeE.ItemInserted:
                ginst = GroupInstance;
                if (!m_bAllowNew || ginst == null) {
                    throw new InvalidOperationException("AddNew is not allowed");
                }
                iBmk = ginst.LineBookmark(ginst.CurrentLine);
                m_ginstRowNewLine = new DataRowGInst(this, iBmk);
                m_arrayList.Insert(iItemIndex, m_ginstRowNewLine);
                OnListChanged(new ListChangedEventArgs(ListChangedType.ItemAdded, iItemIndex));
                break;
            }
        }
    }
}
