using System;
using System.Collections;
using System.IO;
using System.Xml;

namespace RepPP.Toolkit {
	/// <summary>
	/// TypedInstanceGenBase: Base class for typed instance generation
	/// </summary>
	public abstract class TypedInstanceGenBase {
        /// <summary>Prefis for the generated typed card instance class name </summary>
        public const string TYPED_CARD_INST_PREFIX      = "I_";
        /// <summary>Prefix for the generated typed group instance class name </summary>
        public const string TYPED_GROUP_INST_PREFIX     = "I_";
        /// <summary>
        /// Interface of a Log object
        /// </summary>
        public interface ILog {
            /// <summary>
            /// LogClashBetweenGrpAndCrdName:   Log a clash between generated class for a class and a group
            /// </summary>
            /// <param name="crd">              Card</param>
            /// <param name="grp">              Group</param>
            /// <param name="strName">          Name in clash</param>
            void LogClashBetweenGrpAndCrdName(RepPP.Card crd, RepPP.Group grp, string strName);
            /// <summary>
            /// LogClashError:                  Log a clash between a generated property and a base member
            /// </summary>
            /// <param name="strName">          Name in clash</param>
            /// <param name="eBaseMemberType">  Type of the base member</param>
            /// <param name="ePropType">        Type of the property</param>
            void LogClashError(string strName, ClashObjectType eBaseMemberType, ClashObjectType ePropType);
        };
	    /// <summary>Interface to inform project of file update</summary>
	    public interface IProjectFolder {

            //******************************************************************
            //
	        /// <summary>
	        /// Creates a file in the project
	        /// </summary>
	        /// <param name="strFileName">  FileName to add</param>
	        /// <returns>
	        /// Stream to be used to write to the file
	        /// </returns>
	        //
            //******************************************************************
	        StreamWriter CreateFile(string strFileName);

            //******************************************************************
            //
	        /// <summary>
	        /// Open a file for reading
	        /// </summary>
	        /// <param name="strFileName">  Name of the file</param>
	        /// <returns>
	        /// Stream to be used to read the file
	        /// </returns>
	        //
            //******************************************************************
	        StreamReader OpenForReading(string strFileName);

            //******************************************************************
            //
	        /// <summary>
	        /// Open a file for writing
	        /// </summary>
	        /// <param name="strFileName">  Name of the file</param>
	        /// <returns>
	        /// Stream to be used to write to the file
	        /// </returns>
	        //
            //******************************************************************
	        StreamWriter OpenForWriting(string strFileName);

            //******************************************************************
            //
	        /// <summary>
	        /// Delete a file from project
	        /// </summary>
	        /// <param name="strFileName">  Name of the file</param>
	        //
            //******************************************************************
	        void DeleteFile(string strFileName);

            //******************************************************************
            //
	        /// <summary>
	        /// Checks if the file exist in the project folder
	        /// </summary>
	        /// <param name="strFileName">  Name of the file</param>
	        //
            //******************************************************************
	        bool IsFileExistInFolder(string strFileName);

            //******************************************************************
            //
	        /// <summary>
	        /// Checks if the file exist in the project
	        /// </summary>
	        /// <param name="strFileName">  Name of the file</param>
	        //
            //******************************************************************
	        bool IsFileExistInProject(string strFileName);

            //******************************************************************
            //
	        /// <summary>
	        /// Gets the list of files in the project
	        /// </summary>
	        /// <returns>
	        /// List of files
	        /// </returns>
	        //
            //******************************************************************
	        string[] GetFileList();
	    }
        
        /// <summary>Type of member which can cause name clash</summary>
        public enum ClashObjectType {
            /// <summary>Unknwown object type</summary>
            Unknown         = 0,
            /// <summary>Base field (.Net field)</summary>
            BaseField       = 1,
            /// <summary>Base property</summary>
            BaseProperty    = 2,
            /// <summary>Base method</summary>
            BaseMethod      = 3,
            /// <summary>Base constructor</summary>
            BaseCtor        = 4,
            /// <summary>Base event</summary>
            BaseEvent       = 5,
            /// <summary>A property giving access to a child group instance</summary>
            Group           = 6,
            /// <summary>A property giving access to a Rep++ field</summary>
            Field           = 7
        };
	    
	    /// <summary>Name of the configuration file for auto regeneration</summary>
	    private const string            m_strConfigFileName = "TypedConfig.xml";
	    /// <summary>Signature for typed instance files</summary>
	    private const string            m_strSignature = " Rep++ Generated file. Do not modify\r\n\r\n";
	    /// <summary>Web Services generator class</summary>
	    private Web.WebServicesGenBase  m_wsGen;

        //******************************************************************
        //
        /// <summary>
        /// ctor:                               Class constructor
        /// </summary>
        /// <param name="wsGen">                Generator Web Services adapter class</param>
        //
        //******************************************************************
		protected TypedInstanceGenBase(Web.WebServicesGenBase wsGen) {
		    m_wsGen     = wsGen;
		}

        //******************************************************************
        //
        /// <summary>
        /// ctor:                               Class constructor
        /// </summary>
        //
        //******************************************************************
		protected TypedInstanceGenBase() : this(null) {
		}

        //******************************************************************
        //
        /// <summary>
        /// WSGenerator:                        Web services generator if any
        /// </summary>
        //
        //******************************************************************
		protected Web.WebServicesGenBase WSGenerator {
		    get {
		        return(m_wsGen);
		    }
		}

        //******************************************************************
        //
        /// <summary>
        /// GetFieldName:               Gets the name of the a field
        /// </summary>
        /// <param name="fld">          Field object</param>
        /// <returns>
        /// Name of the field
        /// </returns>
        //
        //******************************************************************
        public virtual string GetFieldName(RepPP.Field fld) {
            return(fld.PropertyName);
        }

        //******************************************************************
        //
        /// <summary>
        /// DecorateFieldPropertyName:  Decorate the name of the property for a field
        /// </summary>
        /// <param name="strPrefix">    Prefix to use</param>
        /// <param name="strFldName">   Field name</param>
        /// <returns>
        /// Property name for the field
        /// </returns>
        //
        //******************************************************************
        protected virtual string DecorateFieldPropertyName(string strPrefix, string strFldName) {
            return(strPrefix + strFldName.Replace("$", ""));
        }

        //******************************************************************
        //
        /// <summary>
        /// GetGroupName:               Get the name of a group
        /// </summary>
        /// <param name="grp">          Group</param>
        /// <returns>
        /// Name for the group
        /// </returns>
        //
        //******************************************************************
        public virtual string GetGroupName(RepPP.Group grp) {
            return(grp.ClassName);
        }

        //******************************************************************
        //
        /// <summary>
        /// GetCardGroupName:           Get the name of a card group
        /// </summary>
        /// <param name="cgrp">         Card Group</param>
        /// <returns>
        /// Name of a card group
        /// </returns>
        //
        //******************************************************************
        public virtual string GetCardGroupName(RepPP.CardGroup cgrp) {
            return(cgrp.ClassName);
        }

        //******************************************************************
        //
        /// <summary>
        /// DecorateGroupClassName:     Decorate the name of the generated class for a group
        /// </summary>
        /// <param name="strPrefix">    Prefix to use</param>
        /// <param name="strGrpName">   Group name</param>
        /// <returns>
        /// Class name for the group
        /// </returns>
        //
        //******************************************************************
        protected virtual string DecorateGroupClassName(string strPrefix, string strGrpName) {
            return(strPrefix + strGrpName.Replace("$", ""));
        }

        //******************************************************************
        //
        /// <summary>
        /// DecorateGroupPropertyName:  Decorate the name of the generated property for a group
        /// </summary>
        /// <param name="strPrefix">    Prefix to use</param>
        /// <param name="strGrpName">   Group name</param>
        /// <returns>
        /// Property name for the group
        /// </returns>
        //
        //******************************************************************
        protected virtual string DecorateGroupPropertyName(string strPrefix, string strGrpName) {
            return(strPrefix + strGrpName.Replace("$", ""));
        }

        //******************************************************************
        //
        /// <summary>
        /// GetCardName:                Get the name of a card
        /// </summary>
        /// <param name="crd">          Name of the card</param>
        /// <returns>
        /// Name for the card
        /// </returns>
        //
        //******************************************************************
        public virtual string GetCardName(RepPP.Card crd) {
            return(crd.ClassName);
        }

        //******************************************************************
        //
        /// <summary>
        /// DecorateCardClassName:      Decorates the name of the generated class for a card
        /// </summary>
        /// <param name="strPrefix">    Prefix to use</param>
        /// <param name="strCrdName">   Card name</param>
        /// <returns>
        /// Class name for the card
        /// </returns>
        //
        //******************************************************************
        protected virtual string DecorateCardClassName(string strPrefix, string strCrdName) {
            return(strPrefix + strCrdName.Replace("$", ""));
        }

        //*********************************************************
        //
        /// <summary>
        /// FillHashWithName:       Fills the hash table with base member
        /// </summary>
        /// <param name="hashName"> Hash table to fill</param>
        /// <param name="type">     Type of the class</param>
        //
        //*********************************************************
        private void FillHashWithName(Hashtable hashName, Type type) {
            System.Reflection.MemberInfo[]  arrMI;
            string                          strName;
            ClashObjectType                 eType;
            bool                            bPrivate;

            arrMI = type.GetMembers(System.Reflection.BindingFlags.NonPublic |
                                    System.Reflection.BindingFlags.Instance  |
                                    System.Reflection.BindingFlags.Static);
            foreach (System.Reflection.MemberInfo mi in arrMI) {
                bPrivate = false;
                strName  = mi.Name;
                switch(mi.MemberType) {
                case System.Reflection.MemberTypes.Event:
                    eType    = ClashObjectType.BaseEvent;
                    break;
                case System.Reflection.MemberTypes.Field:
                    bPrivate = ((System.Reflection.FieldInfo)mi).IsPrivate;
                    eType    = ClashObjectType.BaseField;
                    break;
                case System.Reflection.MemberTypes.Constructor:
                    bPrivate = ((System.Reflection.ConstructorInfo)mi).IsPrivate;
                    eType    = ClashObjectType.BaseCtor;
                    break;
                case System.Reflection.MemberTypes.Method:
                    bPrivate = ((System.Reflection.MethodInfo)mi).IsPrivate;
                    eType    = ClashObjectType.BaseMethod;
                    break;
                case System.Reflection.MemberTypes.Property:
                    bPrivate = ((System.Reflection.PropertyInfo)mi).GetGetMethod().IsPrivate &&
                               ((System.Reflection.PropertyInfo)mi).GetSetMethod().IsPrivate;
                    eType    = ClashObjectType.BaseProperty;
                    break;
                default:
                    eType    = ClashObjectType.Unknown;
                    break;
                }
                if (!bPrivate && eType != ClashObjectType.Unknown && !hashName.ContainsKey(strName)) {
                    hashName.Add(strName, eType);
                }
            }
        }

        //*********************************************************
        //
        /// <summary>
        /// CheckGroupNameClash:                Checks for name clash for the specified group
        /// </summary>
        /// <param name="log">                  Log object</param>
        /// <param name="typeGroupBaseClass">   Base class used for generated group instance</param>
        /// <param name="strFldPropPrefix">     Prefix to use for property accessing field instance</param>
        /// <param name="grp">                  Group to be checked</param>
        /// <returns>
        /// true:       Success
        /// false:      Name clash
        /// </returns>
        //
        //*********************************************************
        private bool CheckGroupNameClash(ILog log, Type typeGroupBaseClass, string strFldPropPrefix, RepPP.Group grp) {
            bool        bRetVal = true;
            Hashtable   hashName;
            string      strName;

            hashName = new Hashtable(100);
            FillHashWithName(hashName, typeGroupBaseClass);
            foreach (RepPP.Field fld in grp.Fields) {
                strName = DecorateFieldPropertyName(strFldPropPrefix, GetFieldName(fld));
                if (hashName.ContainsKey(strName)) {
                    log.LogClashError(strName, (ClashObjectType)hashName[strName], ClashObjectType.Field);
                } else {
                    hashName.Add(strName, ClashObjectType.Field);
                }
            }
            return(bRetVal);
        }

        //*********************************************************
        //
        /// <summary>
        /// CheckCardGroupNameClash:            Checks for name clash for the specified group
        /// </summary>
        /// <param name="log">                  Log object</param>
        /// <param name="typeGroupBaseClass">   Base class used for generated group instance</param>
        /// <param name="strFldPropPrefix">     Prefix to use for property accessing field instance</param>
        /// <param name="cgrp">                 Group to be checked</param>
        /// <returns>
        /// true:       Success
        /// false:      Name clash
        /// </returns>
        //
        //*********************************************************
        private bool CheckCardGroupNameClash(ILog log, Type typeGroupBaseClass, string strFldPropPrefix, RepPP.CardGroup cgrp) {
            bool    bRetVal;

            bRetVal = CheckGroupNameClash(log, typeGroupBaseClass, strFldPropPrefix, cgrp.Group);
            foreach (RepPP.CardGroup cgrpChild in cgrp.ChildCardGroups) {
                if (!CheckCardGroupNameClash(log, typeGroupBaseClass, strFldPropPrefix, cgrpChild)) {
                    bRetVal = false;
                }
            }
            return(bRetVal);
        }

        //*********************************************************
        //
        /// <summary>
        /// FillCardGroupList:      Fill a list with all card group in the card hierarchy
        /// </summary>
        /// <param name="cgrp">     Card group</param>
        /// <param name="arrCGrp">  Array to fill</param>
        //
        //*********************************************************
        private void FillCardGroupList(RepPP.CardGroup cgrp, ArrayList arrCGrp) {
            arrCGrp.Add(cgrp);
            foreach (RepPP.CardGroup cgrpChild in cgrp.ChildCardGroups) {
                FillCardGroupList(cgrpChild, arrCGrp);
            }
        }

        //*********************************************************
        //
        /// <summary>
        /// CheckNameForClash:                  Checks if name clash exist with generated field and group name with base class
        /// </summary>
        /// <param name="log">                  Log object</param>
        /// <param name="arrCards">             Array of cards to check</param>
        /// <param name="arrGrps">              Array of groups to check</param>
        /// <param name="typeBaseCardClass">    Type of base class for generated card instance</param>
        /// <param name="typeBaseGroupClass">   Type of base class for generated group instance</param>
        /// <param name="strCrdClsPrefix">      Prefix to use with class name</param>
        /// <param name="strGrpClsPrefix">      Prefix to use with group name</param>
        /// <param name="strGrpPropPrefix">     Prefix to use for property to access group instance</param>
        /// <param name="strFldPropPrefix">     Prefix to use for property to access field instance</param>
        /// <returns>
        /// true:   Ok
        /// false:  Problem. Name clash
        /// </returns>
        /// <remarks>
        /// 1.  Must ensure the name of all classes generated cards and groups don't clash.
        /// 2.  Must ensure that all properties accessing child group instances in card class don't clash with base class.
        /// 3.  Must ensure that all properties for each group class don't clash with base class.
        /// </remarks>
        //
        //*********************************************************        
        public bool CheckForNameClash(ILog          log,
                                      RepPP.Card[]  arrCards,
                                      RepPP.Group[] arrGrps,
                                      Type          typeBaseCardClass,
                                      Type          typeBaseGroupClass,
                                      string        strCrdClsPrefix,
                                      string        strGrpClsPrefix,
                                      string        strGrpPropPrefix,
                                      string        strFldPropPrefix) {
            bool            bRetVal = true;
            RepPP.CardGroup cgrpRoot;
            string          strName;
            Hashtable       hashName;
            ArrayList       arrCGrp;

            hashName = new Hashtable(arrCards.Length);
            foreach (RepPP.Card crd in arrCards) {
                hashName.Add(DecorateCardClassName(strCrdClsPrefix, GetCardName(crd)), crd);
            }
            foreach (RepPP.Group grp in arrGrps) {
                strName = DecorateGroupClassName(strGrpClsPrefix, GetGroupName(grp));
                if (hashName.ContainsKey(strName)) {
                    bRetVal = false;
                    log.LogClashBetweenGrpAndCrdName(hashName[strName] as RepPP.Card, grp, strName);
                }
            }
            if (bRetVal) {
                foreach (RepPP.Card crd in arrCards) {
                    hashName.Clear();
                    cgrpRoot = crd.ChildCardGroups[0];
                    if (cgrpRoot != null) {
                        arrCGrp = new ArrayList(32);
                        FillCardGroupList(cgrpRoot, arrCGrp);
                        FillHashWithName(hashName, typeBaseCardClass);
                        foreach (RepPP.CardGroup cgrp in arrCGrp) {
                            strName = DecorateGroupPropertyName(strGrpPropPrefix, GetCardGroupName(cgrp));
                            if (hashName.ContainsKey(strName)) {
                                log.LogClashError(strName, (ClashObjectType)hashName[strName], ClashObjectType.Group);
                                bRetVal = false;
                            } else {
                                hashName.Add(strName, ClashObjectType.Group);
                            }
                        }
                        if (!CheckCardGroupNameClash(log, typeBaseGroupClass, strFldPropPrefix, cgrpRoot)) {
                            bRetVal = false;
                        }
                    }
                }
                foreach (RepPP.Group grp in arrGrps) {
                    if (!CheckGroupNameClash(log, typeBaseGroupClass, strFldPropPrefix, grp)) {
                        bRetVal = false;
                    }
                }
            }
            return(bRetVal);
        }

        //******************************************************************
        //
        /// <summary>
        /// GenerateTypedGroupInstance:         Generate typed instance for a group
        /// </summary>
        /// <param name="grp">                  Group</param>
        /// <param name="strCrdClassPrefix">    Prefix for generated card class name</param>
        /// <param name="strGrpClassPrefix">    Prefix for generated group class name</param>
        /// <param name="strGrpPropertyPrefix"> Prefix for generated property group name</param>
        /// <param name="strFldPropertyPrefix"> Prefix for generated property field name</param>
        /// <param name="strNameSpace">         The namespace</param>
        /// <param name="strRootNameSpace">     The root namespace of the project</param>
        /// <param name="strBaseClassName">     Name of the base class</param>
        /// <param name="bGenWS">               true to generate Web Services</param>
        /// <returns>
        /// The generated class
        /// </returns>
        //
        //******************************************************************
		public abstract string GenerateTypedGroupInstance(RepPP.Group   grp, 
                                                          string        strCrdClassPrefix,
                                                          string        strGrpClassPrefix,
                                                          string        strGrpPropertyPrefix,
                                                          string        strFldPropertyPrefix,
                                                          string        strNameSpace,
                                                          string        strRootNameSpace,
                                                          string        strBaseClassName,
                                                          bool          bGenWS);

        //******************************************************************
        //
        /// <summary>
        /// GenerateTypedGroupInstance:     Generate typed instance for a group
        /// </summary>
        /// <param name="grp">              Group</param>
        /// <param name="strNameSpace">     The namespace</param>
        /// <param name="strRootNameSpace"> The root namespace of the project</param>
        /// <param name="strBaseClassName"> Name of the base class</param>
        /// <param name="bGenWS">           true to generate Web Services</param>
        /// <returns>
        /// The generated class
        /// </returns>
        //
        //******************************************************************
		public string GenerateTypedGroupInstance(RepPP.Group   grp, 
                                                 string        strNameSpace,
                                                 string        strRootNameSpace,
                                                 string        strBaseClassName,
                                                 bool          bGenWS) {
            return(GenerateTypedGroupInstance(grp,
                                              "CC",
                                              "GC",
                                              "gi",
                                              "fld",
                                              strNameSpace,
                                              strRootNameSpace,
                                              strBaseClassName,
                                              bGenWS));
        }

        //******************************************************************
        //
        /// <summary>
        /// GenerateTypedCardInstance:          Generate typed instance for a card
        /// </summary>
        /// <param name="crd">                  Card</param>
        /// <param name="strCrdClassPrefix">    Prefix for generated card class name</param>
        /// <param name="strGrpClassPrefix">    Prefix for generated group class name</param>
        /// <param name="strGrpPropertyPrefix"> Prefix for generated property group name</param>
        /// <param name="strFldPropertyPrefix"> Prefix for generated property field name</param>
        /// <param name="strNameSpace">         The namespace</param>
        /// <param name="strRootNamespace">     The root namespace of the project</param>
        /// <param name="strCrdBaseClassName">  Name of the base class for the card instance</param>
        /// <param name="strGrpBaseClassName">  Name of the base class for the group instance</param>
        /// <param name="bGenWS">               true to generate Web Services</param>
        /// <returns>
        /// Generated class
        /// </returns>
        //
        //******************************************************************
		public abstract string GenerateTypedCardInstance(RepPP.Card     crd, 
                                                         string         strCrdClassPrefix,
                                                         string         strGrpClassPrefix,
                                                         string         strGrpPropertyPrefix,
                                                         string         strFldPropertyPrefix,
                                                         string         strNameSpace,
                                                         string         strRootNamespace,
                                                         string         strCrdBaseClassName,
                                                         string         strGrpBaseClassName,
                                                         bool           bGenWS);

        //******************************************************************
        //
        /// <summary>
        /// GenerateTypedCardInstance:          Generate typed instance for a card
        /// </summary>
        /// <param name="crd">                  Card</param>
        /// <param name="strNameSpace">         The namespace</param>
        /// <param name="strRootNamespace">     The root namespace of the project</param>
        /// <param name="strCrdBaseClassName">  Name of the base class for the card instance</param>
        /// <param name="strGrpBaseClassName">  Name of the base class for the group instance</param>
        /// <param name="bGenWS">               true to generate Web Services</param>
        /// <returns>
        /// Generated class
        /// </returns>
        //
        //******************************************************************
		public string GenerateTypedCardInstance(RepPP.Card     crd, 
                                                string         strNameSpace,
                                                string         strRootNamespace,
                                                string         strCrdBaseClassName,
                                                string         strGrpBaseClassName,
                                                bool           bGenWS) {
            return(GenerateTypedCardInstance(crd,
                                             "CC",
                                             "GC",
                                             "gi",
                                             "fld",
                                             strNameSpace,
                                             strRootNamespace,
                                             strCrdBaseClassName,
                                             strGrpBaseClassName,
                                             bGenWS));                                                
        }

        //******************************************************************
        //
        /// <summary>
        /// GenerateUserTypedGroupInstance:     Generate user typed instance for a group
        /// </summary>
        /// <param name="grp">                  Group</param>
        /// <param name="strGrpClassPrefix">    Prefix for generated group class name</param>
        /// <param name="strNameSpace">         The namespace</param>
        /// <param name="strRootNameSpace">     The root namespace of the project</param>
        /// <returns>
        /// The generated class
        /// </returns>
        //
        //******************************************************************
		public abstract string GenerateUserTypedGroupInstance(RepPP.Group   grp, 
                                                              string        strGrpClassPrefix,
                                                              string        strNameSpace,
                                                              string        strRootNameSpace);

        //******************************************************************
        //
        /// <summary>
        /// GenerateUserTypedCardInstance:     Generate user typed instance for a card
        /// </summary>
        /// <param name="crd">                  Card</param>
        /// <param name="strCrdClassPrefix">    Prefix for generated card class name</param>
        /// <param name="strNameSpace">         The namespace</param>
        /// <param name="strRootNameSpace">     The root namespace of the project</param>
        /// <returns>
        /// Generated class
        /// </returns>
        //
        //******************************************************************
		public abstract string GenerateUserTypedCardInstance(RepPP.Card     crd, 
                                                             string         strCrdClassPrefix,
                                                             string         strNameSpace,
                                                             string         strRootNameSpace);

        //******************************************************************
        //
		/// <summary>
		/// BuildCardFileName:                  Build the name of a card typed instance
		/// </summary>
        /// <param name="strCrdClassPrefix">    Prefix for generated card class name</param>
		/// <param name="crd">                  Rep++ Card</param>
		/// <param name="strFileExt">           File extension</param>
		/// <returns>
		/// Filename
		/// </returns>
        //
        //******************************************************************
		public virtual string BuildCardFileName(string strCrdClassPrefix, RepPP.Card crd, string strFileExt) {
		    return(DecorateCardClassName(strCrdClassPrefix, (crd == null) ? "" : GetCardName(crd)) + strFileExt);
		}

        //******************************************************************
        //
		/// <summary>
		/// BuildGroupFileName:                 Build the name of a group typed instance
		/// </summary>
        /// <param name="strGrpClassPrefix">    Prefix for generated group class name</param>
		/// <param name="grp">                  Rep++ group</param>
		/// <param name="strFileExt">           File extension</param>
		/// <returns>
		/// Filename
		/// </returns>
        //
        //******************************************************************
		public virtual string BuildGroupFileName(string strGrpClassPrefix, RepPP.Group grp, string strFileExt) {
		    return(DecorateGroupClassName(strGrpClassPrefix, (grp == null) ? "" : GetGroupName(grp)) + strFileExt);
		}

        //******************************************************************
        //
        /// <summary>
        /// GenerateSignature:              Generates the signature
        /// </summary>
        /// <param name="strRemark">        Command for remark (' or //)</param>
        /// <returns>
        /// Signature
        /// </returns>
        //
        //******************************************************************
		protected string GenerateSignature(string strRemark) {
		    return(strRemark + m_strSignature);
		}

        //******************************************************************
        //
		/// <summary>
		/// IsSignaturePresent:             Checks if the signature is present
		/// </summary>
		/// 
		/// <param name="projFolder">       Interface with the project folder</param>
		/// <param name="strFileName">      File name</param>
		/// <returns>
		/// true if succeed, false if not
		/// </returns>
        //
        //******************************************************************
		private bool IsSignaturePresent(IProjectFolder projFolder, string strFileName) {
		    string          strContent;
		    StreamReader    streamReader;
		    
		    streamReader = projFolder.OpenForReading(strFileName);
		    using(streamReader) {
		        strContent = streamReader.ReadToEnd();
		    }
		    return(strContent.IndexOf(m_strSignature) != -1);
		}

        //******************************************************************
        //
		/// <summary>
		/// RegenerateGroups:                   Regenerates all groups which need to be regenerated
		/// </summary>
		/// <param name="app">                  Rep++ application object</param>
		/// <param name="projFolder">           Interface with the project folder</param>
		/// <param name="strFileExt">           File extension (.vb, .cs)</param>
        /// <param name="strCrdClassPrefix">    Prefix for generated card class name</param>
        /// <param name="strGrpClassPrefix">    Prefix for generated group class name</param>
        /// <param name="strGrpPropertyPrefix"> Prefix for generated property group name</param>
        /// <param name="strFldPropertyPrefix"> Prefix for generated property field name</param>
		/// <param name="strNameSpace">         Namespace</param>
        /// <param name="strRootNameSpace">     Root Namespace of the project</param>
		/// <param name="strGroupBaseClass">    Group base class</param>
		/// <param name="bGenWS">               true to generates Web Services structures</param>
		/// <param name="bFullSync">            true to fully synchronize type. false to only sync the existing one</param>
        //
        //******************************************************************
		private void RegenerateGroups(RepPP.Application  app,
		                              IProjectFolder     projFolder,
		                              string             strFileExt,
                                      string             strCrdClassPrefix,
                                      string             strGrpClassPrefix,
                                      string             strGrpPropertyPrefix,
                                      string             strFldPropertyPrefix,
		                              string             strNameSpace,
                                      string             strRootNameSpace,      
		                              string             strGroupBaseClass,
		                              bool               bGenWS,
		                              bool               bFullSync) {
		    bool            bGenerate;
		    bool            bFileExist;
		    string          strFileName;
		    string          strNewContent;
		    string          strOldContent;
		    StreamReader    streamReader;
		    StreamWriter    streamWriter;
            string          strUserFileName;
		    
		    foreach (RepPP.Group grp in app.Groups) {
		        strFileName = BuildGroupFileName(TYPED_GROUP_INST_PREFIX + strGrpClassPrefix, grp, strFileExt);
		        bFileExist  = projFolder.IsFileExistInFolder(strFileName);
		        if (bFileExist || bFullSync) {
		            strNewContent = GenerateTypedGroupInstance(grp,
                                                               strCrdClassPrefix,
                                                               strGrpClassPrefix,
                                                               strGrpPropertyPrefix,
                                                               strFldPropertyPrefix,
		                                                       strNameSpace,
                                                               strRootNameSpace,
		                                                       strGroupBaseClass,
		                                                       bGenWS);
		            if (bFileExist) {
		                using(streamReader = projFolder.OpenForReading(strFileName)) {
		                    strOldContent = streamReader.ReadToEnd();
		                }
		                bGenerate = (strOldContent != strNewContent);
		            } else {
		                bGenerate = bFullSync;
		            }
		            if (bGenerate) {
		                streamWriter = (bFileExist) ? projFolder.OpenForWriting(strFileName) :
		                                              projFolder.CreateFile(strFileName);
                        using(streamWriter) {
                            streamWriter.Write(strNewContent);
                        }
                    }
                    // Generate user typed instance class
                    strUserFileName = BuildGroupFileName(strGrpClassPrefix, grp, strFileExt);
                    if (!projFolder.IsFileExistInProject(strUserFileName)) {
		                strNewContent = GenerateUserTypedGroupInstance(grp,
                                                                       strGrpClassPrefix,
		                                                               strNameSpace,
                                                                       strRootNameSpace);
		                streamWriter = projFolder.CreateFile(strUserFileName);
                        using(streamWriter) {
                            streamWriter.Write(strNewContent);
                        }
                    }
		        }
		    }
		}

        //******************************************************************
        //
		/// <summary>
		/// RegenerateCards:                    Regenerates all cards which need to be regenerated
		/// </summary>
		/// <param name="app">                  Rep++ application object</param>
		/// <param name="projFolder">           Interface with the project folder</param>
		/// <param name="strFileExt">           File extension (.vb, .cs)</param>
        /// <param name="strCrdClassPrefix">    Prefix for generated card class name</param>
        /// <param name="strGrpClassPrefix">    Prefix for generated group class name</param>
        /// <param name="strGrpPropertyPrefix"> Prefix for generated property group name</param>
        /// <param name="strFldPropertyPrefix"> Prefix for generated property field name</param>
		/// <param name="strNameSpace">         Namespace</param>
        /// <param name="strRootNameSpace">     Root namespace of the project</param>
		/// <param name="strCardBaseClass">     Card base class if found</param>
		/// <param name="strGroupBaseClass">    Group base class</param>
		/// <param name="bGenWS">               true to generates Web Services structures</param>
		/// <param name="bFullSync">            true to fully synchronize type. false to only sync the existing one</param>
        //
        //******************************************************************
		private void RegenerateCards(RepPP.Application  app,
		                             IProjectFolder     projFolder,
		                             string             strFileExt,
                                     string             strCrdClassPrefix,
                                     string             strGrpClassPrefix,
                                     string             strGrpPropertyPrefix,
                                     string             strFldPropertyPrefix,
		                             string             strNameSpace,
                                     string             strRootNameSpace,
		                             string             strCardBaseClass,
		                             string             strGroupBaseClass,
		                             bool               bGenWS,
		                             bool               bFullSync) {
		    bool            bGenerate;
		    bool            bFileExist;
		    string          strFileName;
		    string          strNewContent;
		    string          strOldContent;
		    RepPP.Card[]    arrCrd;
		    RepPP.Group[]   arrGrp;
		    StreamReader    streamReader;
		    StreamWriter    streamWriter;
            string          strUserFileName;
		    
		    arrCrd = new RepPP.Card[1];
		    arrGrp = new RepPP.Group[0];
		    foreach (RepPP.Card crd in app.Cards) {
                // Generate typed instance class (_cc)
		        strFileName     = BuildCardFileName(TYPED_CARD_INST_PREFIX + strCrdClassPrefix, crd, strFileExt);
		        bFileExist      = projFolder.IsFileExistInFolder(strFileName);
		        if (bFileExist || bFullSync) {
		            strNewContent = GenerateTypedCardInstance(crd,
                                                              strCrdClassPrefix,
                                                              strGrpClassPrefix,
                                                              strGrpPropertyPrefix,
                                                              strFldPropertyPrefix,
		                                                      strNameSpace,
                                                              strRootNameSpace,
		                                                      strCardBaseClass,
		                                                      strGroupBaseClass,
		                                                      bGenWS);
		            if (bFileExist) {                                          
		                using(streamReader  = projFolder.OpenForReading(strFileName)) {
		                    strOldContent = streamReader.ReadToEnd();
		                }
		                bGenerate = (strOldContent != strNewContent);
		            } else {
		                bGenerate = bFullSync;
		            }
		            if (bGenerate) {
		                streamWriter = (bFileExist) ? projFolder.OpenForWriting(strFileName) :
		                                              projFolder.CreateFile(strFileName);
                        using(streamWriter) {
                            streamWriter.Write(strNewContent);
                        }
                    }
                    // Generate user typed instance class
                    strUserFileName = BuildCardFileName(strCrdClassPrefix, crd, strFileExt);
                    if (!projFolder.IsFileExistInProject(strUserFileName)) {
		                strNewContent = GenerateUserTypedCardInstance(crd,
                                                                      strCrdClassPrefix,
		                                                              strNameSpace,
                                                                      strRootNameSpace);
		                streamWriter = projFolder.CreateFile(strUserFileName);
                        using(streamWriter) {
                            streamWriter.Write(strNewContent);
                        }
                    }
		        }
		    }
		}

        //******************************************************************
        //
		/// <summary>
		/// CheckRegenerateForNameClash:        Checks files that we must regenerate for name clash
		/// </summary>
		/// <param name="app">                  Rep++ application object</param>
        /// <param name="log">                  Logging object</param>
		/// <param name="projFolder">           Interface with the project folder</param>
		/// <param name="strFileExt">           File extension (.vb, .cs)</param>
        /// <param name="strCrdClassPrefix">    Prefix for generated card class name</param>
        /// <param name="strGrpClassPrefix">    Prefix for generated group class name</param>
        /// <param name="strGrpPropertyPrefix"> Prefix for generated property group name</param>
        /// <param name="strFldPropertyPrefix"> Prefix for generated property field name</param>
		/// <param name="bFullSync">            true to fully synchronize type. false to only sync the existing one</param>
        //
        //******************************************************************
		private bool CheckRegenerateForNameClash(RepPP.Application  app,
		                                         ILog               log,
		                                         IProjectFolder     projFolder,
		                                         string             strFileExt,
                                                 string             strCrdClassPrefix,
                                                 string             strGrpClassPrefix,
                                                 string             strGrpPropertyPrefix,
                                                 string             strFldPropertyPrefix,
		                                         bool               bFullSync) {
		    bool            bRetVal;
		    ArrayList       arrCrds;
		    ArrayList       arrGrps;
		    bool            bFileExist;
		    string          strFileName;

		    arrCrds = new ArrayList(app.Cards.Count);
		    arrGrps = new ArrayList(app.Groups.Count);
		    foreach (RepPP.Card crd in app.Cards) {
		        strFileName     = BuildCardFileName(strCrdClassPrefix, crd, strFileExt);
		        bFileExist      = projFolder.IsFileExistInFolder(strFileName);
		        if (bFileExist || bFullSync) {
		            arrCrds.Add(crd);
		        }
		    }
		    foreach (RepPP.Group grp in app.Groups) {
		        strFileName = BuildGroupFileName(strGrpClassPrefix, grp, strFileExt);
		        bFileExist  = projFolder.IsFileExistInFolder(strFileName);
		        if (bFileExist || bFullSync) {
		            arrGrps.Add(grp);
                }		    
		    }
		    bRetVal = CheckForNameClash(log,
		                                (RepPP.Card[])arrCrds.ToArray(typeof(RepPP.Card)),
		                                (RepPP.Group[])arrGrps.ToArray(typeof(RepPP.Group)),
	                                    typeof(RepPP.TypedCardInstanceBase),
	                                    typeof(RepPP.TypedGroupInstanceBase),
	                                    strCrdClassPrefix,
	                                    strGrpClassPrefix,
	                                    strGrpPropertyPrefix,
	                                    strFldPropertyPrefix);
            return(bRetVal);
	    }

        //******************************************************************
        //
		/// <summary>
		/// AddInfo:                Add an information to the config file
		/// </summary>
		/// <param name="xmlNode">  XML node</param>
		/// <param name="strInfo">  Node name</param>
		/// <param name="strVal">   Node value</param>
        //
        //******************************************************************
		private static void AddInfo(XmlNode xmlNode, string strInfo, string strVal) {
		    XmlAttribute    attr;
		    
		    attr        = xmlNode.OwnerDocument.CreateAttribute(strInfo);
		    attr.Value  = strVal;
		    xmlNode.Attributes.Append(attr);
		}

        //******************************************************************
        //
        /// <summary>
        /// GetInfo:                    Gets the information from a config file
        /// </summary>
        /// <param name="xmlNode">      XML node</param>
        /// <param name="strInfo">      Node name</param>
        /// <param name="strDefVal">    Default value</param>
        /// <returns>
        /// Node value
        /// </returns>
        //
        //******************************************************************
		private static string GetInfo(XmlNode xmlNode, string strInfo, string strDefVal) {
		    string      strRetVal;
		    XmlNode     attr;
		    
		    attr      = xmlNode.Attributes.GetNamedItem(strInfo, "");
		    strRetVal = (attr == null) ? strDefVal : attr.Value;
		    return(strRetVal);
		}

        //******************************************************************
        //
        /// <summary>
        /// GetInfo:                Gets the information from a config file
        /// </summary>
        /// <param name="xmlNode">  XML node</param>
        /// <param name="strInfo">  Node name</param>
        /// <returns>
        /// Node value
        /// </returns>
        //
        //******************************************************************
		private static string GetInfo(XmlNode xmlNode, string strInfo) {
		    string      strRetVal;
		    XmlNode     attr;
		    
		    attr      = xmlNode.Attributes.GetNamedItem(strInfo, "");
		    strRetVal = (attr == null) ? "" : attr.Value;
		    return(strRetVal);
		}

        //******************************************************************
        //
		/// <summary>
		/// SaveConfigInfo:                     Save configuration information in the XML file
		/// </summary>
		/// <param name="projFolder">           Interface with the project folder</param>
        /// <param name="strSysName">           System name</param>
        /// <param name="strAppName">           Program name</param>
        /// <param name="strConName">           Connection name</param>
        /// <param name="strUsrName">           User name</param>
        /// <param name="eLangCode">            Language code</param>
        /// <param name="strCrdClassPrefix">    Prefix for generated group class name</param>
        /// <param name="strGrpClassPrefix">    Prefix for generated group class name</param>
        /// <param name="strGrpPropertyPrefix"> Prefix for generated group property name</param>
        /// <param name="strFldPropertyPrefix"> Prefix for generated field property name</param>
		/// <param name="strNameSpace">         Namespace</param>
        /// <param name="strRootNameSpace">     The root namespace</param>
		/// <param name="strCardBaseClass">     Card base class if found</param>
		/// <param name="strGroupBaseClass">    Group base class</param>
		/// <param name="bGenWS">               true to generates Web Services structures</param>
		/// <returns>
		/// true if succeed, false if not
		/// </returns>
        //
        //******************************************************************
		public static bool SaveConfigInfo(IProjectFolder    projFolder,
		                                  string            strSysName,
		                                  string            strAppName,
		                                  string            strConName,
		                                  string            strUsrName,
		                                  LanguageCode      eLangCode,
                                          string            strCrdClassPrefix,
                                          string            strGrpClassPrefix,
                                          string            strGrpPropertyPrefix,
                                          string            strFldPropertyPrefix,
		                                  string            strNameSpace,
                                          string            strRootNameSpace,
		                                  string            strCardBaseClass,
		                                  string            strGroupBaseClass,
		                                  bool              bGenWS) {
		    bool                    bRetVal;
		    XmlDocument             xmlDoc;
		    XmlNode                 xmlNode;
		    StreamWriter            streamWriter;
		    
		    if (strCardBaseClass == null) {
		        strCardBaseClass = String.Empty;
		    }
		    if (strGroupBaseClass == null) {
		        strGroupBaseClass = String.Empty;
		    }
		    streamWriter = projFolder.IsFileExistInFolder(m_strConfigFileName) ?
		                   projFolder.OpenForWriting(m_strConfigFileName) :
		                   projFolder.CreateFile(m_strConfigFileName);
		    try {
		        using(streamWriter) {
		            xmlDoc  = new XmlDocument();
		            xmlNode = xmlDoc.CreateNode(XmlNodeType.Element, "GenInfo", "");
		            xmlDoc.AppendChild(xmlNode);
		            AddInfo(xmlNode, "SystemName",       strSysName);
		            AddInfo(xmlNode, "ApplicationName",  strAppName);
		            AddInfo(xmlNode, "ConnectionName",   strConName);
		            AddInfo(xmlNode, "UserName",         strUsrName);
		            AddInfo(xmlNode, "LanguageCode",     ((int)eLangCode).ToString());
		            AddInfo(xmlNode, "CardClassPrefix",  strCrdClassPrefix);
		            AddInfo(xmlNode, "GroupClassPrefix", strGrpClassPrefix);
		            AddInfo(xmlNode, "GroupPropPrefix",  strGrpPropertyPrefix);
		            AddInfo(xmlNode, "FieldPropPrefix",  strFldPropertyPrefix);
		            AddInfo(xmlNode, "NameSpace",        strNameSpace);
                    AddInfo(xmlNode, "RootNameSpace",    strRootNameSpace);
		            AddInfo(xmlNode, "CardBaseClass",    strCardBaseClass);
		            AddInfo(xmlNode, "GroupBaseClass",   strGroupBaseClass);
		            AddInfo(xmlNode, "GenerateWSStruct", bGenWS.ToString());
		            xmlDoc.Save(streamWriter);
		            bRetVal = true;
		        }
		    } catch {
		        bRetVal = false;
		    }
		    return(bRetVal);
		}

        //******************************************************************
        //
		/// <summary>
		/// ExtractConfigInfo:                  Extract configuration information from the XML file
		/// </summary>
		/// <param name="projFolder">           Interface with the project folder</param>
        /// <param name="strSysName">           System name</param>
        /// <param name="strAppName">           Program name</param>
        /// <param name="strConName">           Connection name</param>
        /// <param name="strUsrName">           User name</param>
        /// <param name="eLangCode">            Language code</param>
        /// <param name="strCrdClassPrefix">    Prefix for generated group class name</param>
        /// <param name="strGrpClassPrefix">    Prefix for generated group class name</param>
        /// <param name="strGrpPropertyPrefix"> Prefix for generated group property name</param>
        /// <param name="strFldPropertyPrefix"> Prefix for generated field property name</param>
		/// <param name="strNameSpace">         Namespace</param>
        /// <param name="strRootNameSpace">     The root namespace of the project</param>
		/// <param name="strCardBaseClass">     Card base class if found</param>
		/// <param name="strGroupBaseClass">    Group base class</param>
		/// <param name="bGenWS">               true to generates Web Services structures</param>
        //
        //******************************************************************
		public static void ExtractConfigInfo(IProjectFolder    projFolder,
		                                     out string        strSysName,
		                                     out string        strAppName,
		                                     out string        strConName,
		                                     out string        strUsrName,
		                                     out LanguageCode  eLangCode,
                                             out string        strCrdClassPrefix,
                                             out string        strGrpClassPrefix,
                                             out string        strGrpPropertyPrefix,
                                             out string        strFldPropertyPrefix,
		                                     out string        strNameSpace,
                                             out string        strRootNameSpace,
		                                     out string        strCardBaseClass,
		                                     out string        strGroupBaseClass,
		                                     out bool          bGenWS) {
		    XmlDocument             xmlDoc;
		    XmlNode                 xmlNode;
		    StreamReader            streamReader;
		    
		    strSysName          = "";
		    strAppName          = "";
		    strConName          = "";
		    strUsrName          = "";
		    eLangCode           = LanguageCode.sdDefault;
		    strRootNameSpace    = "";
            strNameSpace        = "";
		    strCardBaseClass    = "";
		    strGroupBaseClass   = "";
		    bGenWS              = false;
		    if (projFolder.IsFileExistInFolder(m_strConfigFileName)) {
		        using(streamReader = projFolder.OpenForReading(m_strConfigFileName)) {
		            xmlDoc              = new XmlDocument();
		            xmlDoc.Load(streamReader);
		            xmlNode                 = xmlDoc.DocumentElement;
		            strSysName              = GetInfo(xmlNode, "SystemName");
		            strAppName              = GetInfo(xmlNode, "ApplicationName");
		            strConName              = GetInfo(xmlNode, "ConnectionName");
		            strUsrName              = GetInfo(xmlNode, "UserName");
		            eLangCode               = (LanguageCode)Int32.Parse(GetInfo(xmlNode, "LanguageCode"));
		            strCrdClassPrefix       = GetInfo(xmlNode, "CardClassPrefix", "CC");
		            strGrpClassPrefix       = GetInfo(xmlNode, "GroupClassPrefix", "GC");
		            strGrpPropertyPrefix    = GetInfo(xmlNode, "GroupPropPrefix", "gi");
		            strFldPropertyPrefix    = GetInfo(xmlNode, "FieldPropPrefix", "fld");
		            strNameSpace            = GetInfo(xmlNode, "NameSpace");
                    strRootNameSpace        = GetInfo(xmlNode, "RootNameSpace");
		            strCardBaseClass        = GetInfo(xmlNode, "CardBaseClass");
		            strGroupBaseClass       = GetInfo(xmlNode, "GroupBaseClass");
		            bGenWS                  = (GetInfo(xmlNode, "GenerateWSStruct") != "false");
		        }
		    } else {
		        throw new System.ApplicationException("Unable to find the configuration file - " + m_strConfigFileName);
		    }
		}

        //******************************************************************
        //
		/// <summary>
		/// CleanObsoleteTypeFiles:             Clean obsolete file
		/// </summary>
		/// <param name="app">                  Rep++ application object</param>
		/// <param name="projFolder">           Interface with the project folder</param>
        /// <param name="strFileExt">           File extension</param>
        /// <param name="strCrdClassPrefix">    Prefix for generated group class name</param>
        /// <param name="strGrpClassPrefix">    Prefix for generated group class name</param>
        //
        //******************************************************************
		private void CleanObsoleteTypeFiles(RepPP.Application   app,
		                                    IProjectFolder      projFolder,
		                                    string              strFileExt,
                                            string              strCrdClassPrefix,
                                            string              strGrpClassPrefix) {
		    string[]    arrFileName;
		    string      strCardFileName;
		    string      strGroupFileName;
		    string      str;

		    strCardFileName     = BuildCardFileName(strCrdClassPrefix, null, strFileExt).ToUpper();
		    strGroupFileName    = BuildGroupFileName(strGrpClassPrefix, null, strFileExt).ToUpper();
		    arrFileName         = projFolder.GetFileList();
		    foreach (string strFileName in arrFileName) {
		        str = strFileName.ToUpper();
		        if (str.EndsWith(strCardFileName)) {
		            if (IsSignaturePresent(projFolder, strFileName)) {
		                str = str.Substring(0, str.LastIndexOf(strCardFileName));
		                if (app.Cards[str] == null) {
		                    projFolder.DeleteFile(strFileName);
		                }
		            }
		        } else if (str.EndsWith(strGroupFileName)) {
		            if (IsSignaturePresent(projFolder, strFileName)) {
		                str = str.Substring(0, str.LastIndexOf(strGroupFileName));
		                if (app.Groups[str] == null) {
		                    projFolder.DeleteFile(strFileName);
		                }
		            }
		        }
		    }
		}

        //******************************************************************
        //
		/// <summary>
		/// RegenerateTypedInstance:            Regenerates all cards and groups which need to be regenerated
		/// </summary>
		/// <param name="app">                  Rep++ application object</param>
        /// <param name="log">                  Log object for name clash</param>
		/// <param name="projFolder">           Interface with the project folder</param>
		/// <param name="strFileExt">           File extension (.vb, .cs)</param>
		/// <param name="bFullSync">            true to fully synchronize type. false to only sync the existing one</param>
        /// <returns>
        /// true if succeed, false if failed
        /// </returns>
        //
        //******************************************************************
		protected virtual bool RegenerateTypedInstance(RepPP.Application app, ILog log, IProjectFolder projFolder, string strFileExt, bool bFullSync) {
		    bool            bRetVal = true;
		    string          strSysName;
		    string          strAppName;
		    string          strConName;
		    string          strUsrName;
		    LanguageCode    eLangCode;
            string          strCrdClassPrefix;
            string          strGrpClassPrefix;
            string          strGrpPropertyPrefix;
            string          strFldPropertyPrefix;
		    string          strNameSpace;
            string          strRootNameSpace;
		    string          strCardBaseClass;
		    string          strGroupBaseClass;
		    bool            bGenWS;
		    
		    ExtractConfigInfo(projFolder,
		                      out strSysName,
		                      out strAppName,
		                      out strConName,
		                      out strUsrName,
		                      out eLangCode,
                              out strCrdClassPrefix,
                              out strGrpClassPrefix,
                              out strGrpPropertyPrefix,
                              out strFldPropertyPrefix,
		                      out strNameSpace,
                              out strRootNameSpace,
		                      out strCardBaseClass,
		                      out strGroupBaseClass,
		                      out bGenWS);
            if (strNameSpace == String.Empty && strRootNameSpace == String.Empty) {
                System.Windows.Forms.MessageBox.Show("The project does not have a root or a default namespace and no namespace is specified for the typed instances. \r\nPlease specify a namespace for the typed instances.",
                                                     "Rep++ Typed Instances Add-In",
                                                     System.Windows.Forms.MessageBoxButtons.OK,
                                                     System.Windows.Forms.MessageBoxIcon.Error);
            } else {
		        if (app == null) {
                    app = RepPP.Application.GetApplication();
                    if (app != null) {
                         RepPP.Application.Reset(null);
                    }
                    app = RepPP.Application.GetApplicationForDesign(null);
		            if (app == null) {
		                throw new System.ApplicationException("Unable to initialize Rep++");
		            }
		        }
		        if (log != null) {
		            bRetVal = CheckRegenerateForNameClash(app,
		                                                  log,
		                                                  projFolder,
		                                                  strFileExt,
		                                                  strCrdClassPrefix,
		                                                  strGrpClassPrefix,
		                                                  strGrpPropertyPrefix,
		                                                  strFldPropertyPrefix,
		                                                  bFullSync);
		        }
		        if (bRetVal) {
	                RegenerateCards(app,
	                                projFolder,
	                                strFileExt,
                                    strCrdClassPrefix,
                                    strGrpClassPrefix,
                                    strGrpPropertyPrefix,
                                    strFldPropertyPrefix,
                                    strNameSpace,
                                    strRootNameSpace,
	                                strCardBaseClass,
	                                strGroupBaseClass,
	                                bGenWS,
	                                bFullSync);
	                RegenerateGroups(app,
	                                 projFolder,
	                                 strFileExt,
                                     strCrdClassPrefix,
                                     strGrpClassPrefix,
                                     strGrpPropertyPrefix,
                                     strFldPropertyPrefix,
	                                 strNameSpace,
                                     strRootNameSpace,
	                                 strGroupBaseClass,
	                                 bGenWS,
	                                 bFullSync);
                    if (bFullSync) {		                         
	                    CleanObsoleteTypeFiles(app, projFolder, strFileExt, strCrdClassPrefix, strGrpClassPrefix);
	                }
	            }
            }
            return(bRetVal);
		}

        //******************************************************************
        //
		/// <summary>
		/// RegenerateTypedInstance:            Regenerates all cards and groups which need to be regenerated
		/// </summary>
		/// <param name="app">                  Rep++ application object</param>
        /// <param name="log">                  Log object for name clash</param>
		/// <param name="projFolder">           Interface with the project folder</param>
		/// <param name="bFullSync">            true to fully synchronize type. false to only sync the existing one</param>
        //
        //******************************************************************
		public abstract void RegenerateTypedInstance(RepPP.Application app, ILog log, IProjectFolder projFolder, bool bFullSync);

        //******************************************************************
        //
		/// <summary>
		/// RegenerateTypedInstance:            Regenerates all cards and groups which need to be regenerated
		/// </summary>
        /// <param name="log">                  Log object for name clash</param>
		/// <param name="projFolder">           Interface with the project folder</param>
		/// <param name="bFullSync">            true to fully synchronize type. false to only sync the existing one</param>
        //
        //******************************************************************
		public abstract void RegenerateTypedInstance(ILog log, IProjectFolder projFolder, bool bFullSync);
	} // class TypedInstanceGenBase
} // namespace
