using System;
using System.Collections;

namespace RepPP.Toolkit.Window {
	/// <summary>
	/// Summary description for CmdHolder.
	/// </summary>     
	public class CmdHolder {
	    /// <summary>Delegate for the command processing</summary>
        public delegate void    OnCommand();                       // Pointer to the command implementation
	    
        /// <summary>State of the trigger attached to the command</summary>
        public enum CmdStateE {
            /// <summary>Unknown state</summary>
            None        =   0,                                      // Initialization
            /// <summary>Trigger is hidden</summary>
            Hidden      =   1, 
            /// <summary>Trigger is disabled</summary>                                     
            Disabled    =   2,
            /// <summary>Trigger is enabled and visibled</summary>
            Enabled     =   3
        };

        /// <summary>
        /// Array of commands
        /// </summary>
        protected ArrayList               m_arrCmd;

        //*********************************************************
        //
        /// <summary>
        /// ctor:                           Class constructor
        /// </summary>
	    //
        //*********************************************************
		public CmdHolder() {
		    m_arrCmd = new ArrayList();
		}

        //*********************************************************
        //
		/// <summary>
		/// FindCommand:                   Find a command using its name
		/// </summary>
		/// <param name="strCmdName">       Command name</param>
		/// <returns>
		/// Command or null if not found
		/// </returns>
	    //
        //*********************************************************
		public Cmd FindCommand(string strCmdName) {
		    Cmd     cmdRetVal = null;
		    
		    foreach (Cmd cmd in m_arrCmd) {
		        if (cmd.m_strCmdName == strCmdName) {
		            cmdRetVal = cmd;
		            break;
		        }
		    }
		    return(cmdRetVal);
		}

        //*********************************************************
        //
		/// <summary>
		/// AddCmd:                         Add a new command.
		/// </summary>
		/// <param name="strCmdName">       Command name</param>
		/// <param name="del">              Delegate to execute the command</param>
		/// <param name="eState">           Original state</param>
	    //
        //*********************************************************
		public virtual void AddCmd(string strCmdName, OnCommand del, CmdStateE eState) {
		    if (FindCommand(strCmdName) != null) {
		        throw new System.ApplicationException("The command '" + strCmdName + "' already defined");
		    }
		    CheckState(strCmdName, ref eState);
		    m_arrCmd.Add(new Cmd(strCmdName, del, eState));
		}

        //*********************************************************
        //
		/// <summary>
		/// AddCmd:                         Add a new command.
		/// </summary>
		/// <param name="cmd">              Command object</param>
	    //
        //*********************************************************
		public void AddCmd(Cmd cmd) {
		    if (FindCommand(cmd.Name) != null) {
		        throw new System.ApplicationException("The command '" + cmd.Name + "' already defined");
		    }
		    CheckState(cmd.Name, ref cmd.m_eState);
		    m_arrCmd.Add(cmd);
		}

        //*********************************************************
        //
		/// <summary>
		/// RemoveCmd:                      Remove a command.
		/// </summary>
		/// <param name="strCmdName">       Command to remove</param>
	    //
        //*********************************************************
		public void RemoveCmd(string strCmdName) {
		    Cmd   cmd;
		    
		    cmd = FindCommand(strCmdName);
		    if (cmd != null) {
		        m_arrCmd.Remove(cmd);
		        cmd.Dispose();
		    }
		}

        //*********************************************************
        //
        /// <summary>
        /// AddTrigger:                     Add trigger to the command
        /// </summary>
        /// <param name="app">              Rep++ Application if want to set the trigger's text</param>
        /// <param name="strCmdName">       Command name</param>
        /// <param name="arrTrigger">       Array of trigger</param>
	    //
        //*********************************************************
		public void AddTrigger(RepPP.IApplication app, string strCmdName, object[] arrTrigger) {
		    Cmd     cmd;
		    string  strText;
		    
		    cmd = FindCommand(strCmdName);
		    if (cmd == null) {
		        throw new System.ApplicationException("The command '" + strCmdName + "' is not defined");
		    }
		    cmd.AddTrigger(arrTrigger);
		    if (app != null) {
                if (app.IsAtomExist("cmd" + strCmdName, RepPP.AtomType.sdAtomMsg)) {
		            strText = app.GetAtomText("cmd" + strCmdName, RepPP.AtomType.sdAtomMsg);
		            if (strText != null && strText != "") {
		                foreach (object obj in arrTrigger) {
		                    cmd.SetText(obj, strText);
		                }
		            }
                }
		    }
		}

        //*********************************************************
        //
        /// <summary>
        /// RemoveTrigger:                  Remove trigger from the command
        /// </summary>
        /// <param name="strCmdName">       Command name</param>
        /// <param name="arrTrigger">       Array of trigger</param>
        //
        //*********************************************************
        public void RemoveTrigger(string strCmdName, object[] arrTrigger) {
            Cmd     cmd;
		    
            cmd = FindCommand(strCmdName);
            if (cmd == null) {
                throw new System.ApplicationException("The command '" + strCmdName + "' is not defined");
            }
            foreach (object obj in arrTrigger) {
                cmd.RemoveTrigger(obj);
            }
        }

        //*********************************************************
        //
		/// <summary>
		/// ExecuteCmd:                     Execute a command
		/// </summary>
		/// <param name="sender">           Trigger sending the command</param>
		/// <returns>
		/// true if trigger found, false if not.
		/// </returns>
	    //
        //*********************************************************
		public bool ExecuteCmd(object sender) {
		    bool    bRetVal = false;
		    Cmd     cmdToExecute = null;
		    
		    foreach (Cmd cmd in m_arrCmd) {
		        foreach (object obj in cmd.m_arrTrigger) {
		            if (obj == sender) {
		                cmdToExecute = cmd;
		                break;
		            }
		        }
		        if (cmdToExecute != null) {
		            break;
		        }
		    }
		    if (cmdToExecute != null) {
		        bRetVal = true;
		        if (cmdToExecute.m_eState == CmdStateE.Enabled) {
		            cmdToExecute.m_del();
		        }
		    }
		    return(bRetVal);
		}

        //*********************************************************
        //
		/// <summary>
		/// CheckState:                     Check if the state meet security or other requirement
		/// </summary>
		/// <param name="strCmdName">       Command name</param>
		/// <param name="eState">           State of the command</param>
	    //
        //*********************************************************
		public virtual void CheckState(string strCmdName, ref CmdStateE eState) {
		}

        //*********************************************************
        //
		/// <summary>
		/// SetCmdState:                    Set the state of a command
		/// </summary>
		/// <param name="strCmdName">       Command name</param>
		/// <param name="eState">           New state of the command</param>
	    //
        //*********************************************************
		public void SetCmdState(string strCmdName, CmdStateE eState) {
		    Cmd     cmd;
		    
		    cmd = FindCommand(strCmdName);
		    if (cmd == null) {
		        throw new System.ApplicationException("Command '" + strCmdName + "' is not defined");
		    }
		    CheckState(strCmdName, ref eState);
		    cmd.SetState(eState);
		}

        //*********************************************************
        //
		/// <summary>
		/// GetCmdState:                    Gets the state of a command
		/// </summary>
		/// <param name="strCmdName">       Command name</param>
        /// <returns>
        /// State of the command
        /// </returns>
	    //
        //*********************************************************
		public CmdStateE GetCmdState(string strCmdName) {
		    Cmd     cmd;
		    
		    cmd = FindCommand(strCmdName);
		    if (cmd == null) {
		        throw new System.ApplicationException("Command '" + strCmdName + "' is not defined");
		    }
		    return(cmd.State);
		}

        //*********************************************************
        //
        /// <summary>
        /// SetCmdDelegate:                 Set the command delegate
        /// </summary>
        /// <param name="strCmdName">       Command name</param>
        /// <param name="del">              New command delegate</param>
        //
        //*********************************************************
        public void SetCmdDelegate(string strCmdName, CmdHolder.OnCommand del) {
            Cmd     cmd;
		    
            cmd = FindCommand(strCmdName);
            if (cmd == null) {
                throw new System.ApplicationException("Command '" + strCmdName + "' is not defined");
            }
            cmd.m_del = del;
        }
	}
}
